import os
import importlib
from atradebot.strategies.strategy import Strategy

__MODEL_REGISTRY__ = {}

def get_strategy(name:str):
    return __MODEL_REGISTRY__[name]

def register_strategy(name:str):
    def register_model_fn(fn):
        if name in __MODEL_REGISTRY__:
            return __MODEL_REGISTRY__[name]
        if not issubclass(fn, Strategy):
            raise ValueError(f"Model ({name}: {fn.__name__}) must be a subclass of Strategy")
        __MODEL_REGISTRY__[name] = fn
        return fn
    return register_model_fn

for file in os.listdir(os.path.dirname(__file__)):
    if file.endswith('.py') and not file.startswith('__'):
        model_name = file[:file.find('.py')]
        module = importlib.import_module(f"atradebot.strategies.{model_name}")
        





import pandas as pd
from argparse import ArgumentParser
import os
from datetime import date, datetime, timedelta
import shutil
from atradebot.params import DATE_FORMAT, ROOT_D, logging
from atradebot.strategies import __MODEL_REGISTRY__
from atradebot.utils.backtest import Backtester  
from atradebot.utils.utils import find_business_day, gen_report
from atradebot.utils.search import findstocks
from atradebot.strategies.strategy import StrategyConfig
from atradebot.utils.utils import ALLTICKS
from atradebot.utils.alpaca_trade import alpaca_execute

def auto_trade(args, portfolio_path):
    """
    Execute the main strategy based on the given configuration.

    Args:
        config (StrategyConfig): An object containing the necessary configuration parameters.
    
    Returns:
        None
    """    
    args.present_date = date.today()
    args.past_date = find_business_day(args.present_date, -args.inference_days)
    backtester = Backtester(portfolio_path, args=args, zero=False)   
    action = backtester.strategy.generate_allocation()
    alpaca_execute(action)
    #update portfolio
    backtester.portfolio.execute(action)
    backtester.portfolio.portfolio.to_csv(portfolio_path, index=False) # update portfolio
    


def get_args(raw_args=None):
    """
    Get command line arguments for the script.

    Args:
        raw_args (list): A list of raw arguments from the command line. Default is None.

    Returns:
        argparse.Namespace: An object containing the parsed arguments.

    Raises:
        None.
    """    
    parser = ArgumentParser(description="parameters")
    parser.add_argument('--strategy','-s', type=str, default='GPTStrategy', help=f'Modes: {__MODEL_REGISTRY__.keys()}')
    parser.add_argument('--inference_days','-i', type=int, default=20, help='number of days for inference')
    parser.add_argument('--future_days','-f', type=int, default=10, help='number of days in the future to predict')
    parser.add_argument('--newsapi','-n', type=str, default="finhub", help='newsapi: finhub, newsapi, google')
    args = parser.parse_args(raw_args)
    return args
# modes: SimpleStrategy, FinLlama3Strategy, GPTStrategy, ARIMAStrategy 

if __name__ == "__main__":
    args = get_args()
    print(args)
    assert args.strategy != 'ManualStrategy', f"Invalid strategy: {args.strategy}"
    config = StrategyConfig(
        inference_days=args.inference_days,
        future_days=args.future_days,
        strategy=args.strategy,
        newsapi=args.newsapi,
    )
    portfolio_path = f'{ROOT_D}/sd/test_portfolio.csv'
    auto_trade(config, portfolio_path)


# finetune llama-3 https://medium.com/@alexandros_chariton/how-to-fine-tune-llama-3-2-instruct-on-your-own-data-a-detailed-guide-e5f522f397d7
# https://github.com/AI4Finance-Foundation/FinGPT/tree/master/fingpt/FinGPT_Forecaster


from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    BitsAndBytesConfig,
    TrainingArguments,
    Trainer,
    logging,
)
from peft import (
    LoraConfig,
    PeftModel,
    prepare_model_for_kbit_training,
    get_peft_model,
)
import os, torch
from datasets import load_dataset, Dataset
from argparse import ArgumentParser
import bitsandbytes as bnb
from functools import partial
from atradebot.data.llm_prompt import format_chat_template

torch_dtype = torch.float16
attn_implementation = "eager"
BASEMODEL = "meta-llama/Llama-3.2-3B-Instruct"
LORAMODEL = "achang/llama-3.2-3b_lora"
DATASET_NAME = "achang/test_stock"
RESPONSE_TOKENS = '<|start_header_id|>assistant<|end_header_id|>\n\n'

def find_all_linear_names(model):
    cls = bnb.nn.Linear4bit
    lora_module_names = set()
    for name, module in model.named_modules():
        if isinstance(module, cls):
            names = name.split('.')
            lora_module_names.add(names[0] if len(names) == 1 else names[-1])
    if 'lm_head' in lora_module_names:  # needed for 16 bit
        lora_module_names.remove('lm_head')
        return list(lora_module_names)

def tokenize_function(max_length, tokenizer, example):
    
    tgt_id = example['text'].find(RESPONSE_TOKENS)
    prompt_ids = tokenizer.encode(example['text'][:tgt_id].strip(), padding=False,
        max_length=max_length, truncation=True)
    
    target_ids = tokenizer.encode(example['text'][tgt_id:].strip(), padding=False, 
                                  max_length=max_length, truncation=True, add_special_tokens=False)
    
    input_ids = prompt_ids + target_ids
    exceed_max_length = len(input_ids) >= max_length
    if input_ids[-1] != tokenizer.eos_token_id and not exceed_max_length: # Add EOS Token
        input_ids.append(tokenizer.eos_token_id)
    label_ids = [tokenizer.pad_token_id] * len(prompt_ids) + input_ids[len(prompt_ids):]
    return {
        "input_ids": input_ids,
        "labels": label_ids,
        "exceed_max_length": exceed_max_length
    }

def train_model(args):
    # QLoRA config
    bnb_config = BitsAndBytesConfig(
        load_in_4bit=True,
        bnb_4bit_quant_type="nf4",
        bnb_4bit_compute_dtype=torch_dtype,
        bnb_4bit_use_double_quant=True,
    )
    # Load model
    model = AutoModelForCausalLM.from_pretrained(
        BASEMODEL,
        quantization_config=bnb_config,
        device_map="auto",
        attn_implementation=attn_implementation
    )

    # Load tokenizer
    tokenizer = AutoTokenizer.from_pretrained(BASEMODEL, trust_remote_code=True)
    tokenizer.pad_token = tokenizer.eos_token

    # Dataset 
    dataset = load_dataset(DATASET_NAME, split="train")
    dataset = dataset.shuffle(seed=65) # Only use 1000 samples for quick demo
    dataset = dataset.train_test_split(test_size=0.2)
    dataset = dataset.map(partial(format_chat_template, tokenizer),num_proc= 4,)

    # Tokenize the data
    tokenized_dataset = dataset.map(partial(tokenize_function, args.max_length, tokenizer))
    tokenized_dataset = tokenized_dataset.filter(lambda x: not x['exceed_max_length'])
    tokenized_dataset = tokenized_dataset.remove_columns(['instruction', 'response', 'text', 'exceed_max_length'])

    modules = find_all_linear_names(model)
    # LoRA config #TODO: https://arxiv.org/pdf/2501.06252 SVF is better than lora
    peft_config = LoraConfig(
        r=16,
        lora_alpha=32,
        lora_dropout=0.05,
        bias="none",
        task_type="CAUSAL_LM",
        target_modules=modules
    )
    model = get_peft_model(model, peft_config)
    #Hyperparamter   
    training_arguments = TrainingArguments(
        output_dir='./output_dir',
        per_device_train_batch_size=1,
        per_device_eval_batch_size=1,
        gradient_accumulation_steps=2,
        optim="paged_adamw_32bit",
        num_train_epochs=1,
        eval_strategy="steps",
        eval_steps=0.2,
        warmup_steps=5,
        logging_strategy="steps",
        learning_rate=2e-4,
        max_grad_norm=2,
        fp16=True,
        report_to="tensorboard"
    )

    trainer = Trainer(
        model=model,
        args=training_arguments,
        train_dataset=tokenized_dataset["train"],
        eval_dataset=tokenized_dataset["test"],
        tokenizer=tokenizer)

    trainer.train()
    trainer.model.push_to_hub(LORAMODEL)


def test_model(args):
    model = AutoModelForCausalLM.from_pretrained(
            BASEMODEL,
            trust_remote_code=True, 
            device_map="auto",
            torch_dtype=torch.float16,
    )
    # model = PeftModel.from_pretrained(bmodel, new_model) # TODO: maybe dont need train
    model = model.eval()
    tokenizer = AutoTokenizer.from_pretrained(BASEMODEL)
    dataset = load_dataset(DATASET_NAME, split="train").select(range(10))
    dataset = dataset.map(partial(format_chat_template, tokenizer), num_proc= 4)

    prompt = dataset[0]["text"]
    ii = prompt.find(RESPONSE_TOKENS)
    prompt = prompt[:ii]
    inputs = tokenizer(prompt, return_tensors='pt', padding=False)
    inputs = {key: value.to(model.device) for key, value in inputs.items()}
    
    res = model.generate(
            **inputs, max_length=4096, do_sample=True,
            eos_token_id=tokenizer.eos_token_id, use_cache=True)
    output = tokenizer.decode(res[0], skip_special_tokens=True)
    print(prompt)
    print("Output: ======================")
    print(output)

    

def get_parser(raw_args=None):
    parser = ArgumentParser(description="train model")
    parser.add_argument('--mode', type=str, default='eval',
                        help='train or eval')   
    parser.add_argument("--max_length", default=1024, type=int)
    args = parser.parse_args(raw_args)
    return args

if __name__ == "__main__":
    args = get_parser()
    if args.mode == 'train':
        train_model(args)
    else:
        test_model(args)


# get historical OHLCV data from yfinance
import pandas as pd
import os
from atradebot.params import YFCACHE, FHUB_API
import yfinance as yf
import finnhub
from tenacity import retry, stop_after_attempt, wait_fixed

yf.set_tz_cache_location(f"{YFCACHE}/.yfcache")

finnhub_client = finnhub.Client(api_key=FHUB_API)

YFDETAIL_KEYS = ['beta', 'forwardPE', 'volume', 'regularMarketVolume', 'averageVolume', 'averageVolume10days', 'averageDailyVolume10Day',
            'marketCap', 'fiftyTwoWeekLow', 'fiftyTwoWeekHigh', 'priceToSalesTrailing12Months', 'fiftyDayAverage', 'twoHundredDayAverage', 'enterpriseValue', 'profitMargins',
            'floatShares', 'sharesOutstanding', 'sharesShort', 'sharesShortPriorMonth','heldPercentInsiders', 'heldPercentInstitutions', 'shortRatio','impliedSharesOutstanding', 'bookValue', 'trailingEps', 'forwardEps','52WeekChange', 'targetHighPrice', 'targetLowPrice', 'targetMeanPrice',
            'targetMedianPrice', 'enterpriseToRevenue', 'enterpriseToEbitda','totalCash', 'totalCashPerShare', 'ebitda', 'totalDebt', 'quickRatio','currentRatio', 'totalRevenue', 'revenuePerShare', 'returnOnAssets',
            'grossProfits', 'freeCashflow', 'operatingCashflow', 'revenueGrowth','grossMargins', 'ebitdaMargins', 'operatingMargins']


def get_yf_detail(ticker:str) -> str:
    company = yf.Ticker(ticker)
    longBusinessSummary = company.info.get('longBusinessSummary', 'unknown')
    industry = company.info.get('industry', 'unknown')
    country = company.info.get('country', 'unknown')
    company_text = f"Description: {longBusinessSummary}, Industry: {industry}, Location: {country}"
    officers_text = ""
    for officer in company.info['companyOfficers']:
        officers_text += f"{officer['name']}, {officer['title']} \n"
    fullTimeEmployees = company.info.get('fullTimeEmployees', 'unknown')
    company_person = f"Number of employees: {fullTimeEmployees}, Officers: {officers_text} \n"
    sc = [f"{k}: {v}" for k, v in company.info.items() if k in YFDETAIL_KEYS]
    company_details = ', '.join(sc)
    
    return company.info, company_text, company_person, company_details

def get_yf_data_all(ticker:str) -> pd.DataFrame:
    """
    Get all data from yfinance for a stock, with a fallback to Polygon.

    Args:
        ticker (str): Stock ticker.

    Returns:
        pd.DataFrame: Data for the stock.
    """    
    try:
        # Try to get data from yfinance
        data = yf.download(ticker, period="max", interval="1d")
    except Exception as e:
        print(f"yfinance data retrieval failed: {e}.")
    return data

def get_yf_data_some(ticker:str, past_date:str, present_date:str) -> pd.DataFrame:
    """
    Retrieve Yahoo Finance data for a specific ticker within a date range, with a fallback to Polygon.

    Args:
        ticker (str): The ticker symbol of the stock.
        past_date (str): The start date in YYYY-MM-DD format.
        present_date (str): The end date in YYYY-MM-DD format.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the stock data.
    """    
    try:
        # Try to get data from yfinance
        data = yf.download(ticker, start=past_date, end=present_date, interval="1d")
    except Exception as e:
        print(f"yfinance data retrieval failed: {e}.")
    return data

def get_price(stock):
    """
    Get current price of stock.

    Args:
        stock (str): Stock id.

    Returns:
        float: Current price.
    """    
    ticker = yf.Ticker(stock).info
    return ticker['regularMarketOpen']

def get_price_date(date, present_date, stock):
    """
    Get the price of a stock for a given date range.

    Args:
        date (str): Start date to get the stock price.
        present_date (str): End date to get the stock price.
        stock (str): Stock ID.
    
    Returns:
        pandas.DataFrame: Dataframe of stock data from the date to the present_date.
    """    
    data = yf.Ticker(stock)
    hdata = data.history(start=date,  end=present_date)
    return hdata

def get_finhub_data(ticker: str) -> dict:
    """
    Get stock data from Finnhub.

    Args:
        ticker (str): Stock ticker.

    Returns:
        dict: Data for the stock.
    """
    data = finnhub_client.company_profile2(symbol=ticker)
    return data

def get_finhub_price(ticker: str) -> float:
    """
    Get current price of stock from Finnhub.

    Args:
        ticker (str): Stock ticker.

    Returns:
        float: Current price.
    """
    quote = finnhub_client.quote(symbol=ticker)
    return quote['c']



# project global defines
import os
import diskcache as dc
import json
import logging
import git

def get_git_root(path):
    git_repo = git.Repo(path, search_parent_directories=True)
    git_root = git_repo.git.rev_parse("--show-toplevel")
    return git_root


DATE_FORMAT = "%Y-%m-%d" # same as yfinance

ROOT_DIR = os.path.abspath(__file__)
ROOT_D = get_git_root(ROOT_DIR) + '/'

OUTPUT_D = f'{ROOT_D}/sd/output/'
if not os.path.exists(OUTPUT_D):
    os.mkdir(OUTPUT_D)

# Data
HF_LOCALDATA = f'{ROOT_D}/sd/atradebot_collect_' # collect to add to dataset (cronjob)
HF_DATASET = 'achang/atradebot_dataset' # dataset for training
NEWSCACHE_DIR = f'{ROOT_D}/sd/news_cache'
NEWSCACHE = dc.Cache(NEWSCACHE_DIR) # cache news data for inference 

YFCACHE = f'{ROOT_D}/sd/' # cache yfinance data for inference
DATA_KEYS = ['Date', 'News', 'Tickers'] #must have keys in dataset

# Configure logging
logging.basicConfig(
    level=logging.INFO,  # Set log level
    format="%(asctime)s - %(levelname)s - %(message)s",
    handlers=[
        logging.FileHandler(f"{ROOT_D}/sd/app.log"),  # Log to file
        logging.StreamHandler()  # Print to terminal
    ]
)
# List of tickers
# get all stock tickers from https://www.sec.gov/files/company_tickers.json
TICKERS = f'{ROOT_D}/sd/company_tickers.json' # list of stock tickers
with open(TICKERS, 'r') as f:
    alltickers = json.load(f)
ALLTICKS = [v['ticker'] for _, v in alltickers.items()]
rmticks = ['AI', 'IT', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'AR', 'VR']
ALLTICKS = [t for t in ALLTICKS if t not in rmticks]

#APIs
FHUB_API = os.environ.get('FINNHUB_API_KEY') # for getting news finhub
NEWS_API = os.environ.get('NEWS_API_KEY')
OPENAI_API = os.environ.get('OPENAI_API_KEY') # for LLM

ALPACA_API = os.environ.get('ALPACA_API_KEY') # trading platform
ALPACA_SECRET = os.environ.get('ALPACA_SECRET_KEY') 
PAPERTRADE = True # paper trading or live trading


import pandas as pd
from statsmodels.tsa.arima.model import ARIMA
from atradebot.data.yf_data import get_yf_data_some
from atradebot.strategies import register_strategy
from atradebot.strategies.strategy import Strategy, StrategyConfig
import os
from atradebot.utils.utils import find_business_day
import pickle
from pmdarima import auto_arima

@register_strategy("ARIMAStrategy")
class ARIMAStrategy(Strategy):
    """
    A class representing an ARIMA strategy for stock trading.

    Attributes:
        args (StrategyConfig): The configuration parameters for the strategy.
        train_past_date (datetime): The past date used for training the ARIMA models.
        stock_data (dict): A dictionary containing stock price data for each stock in the strategy's portfolio.
    
    Methods:
        predict(present_date, future_days, stock): Predicts the future stock prices using ARIMA models.
        generate_allocation(date, portfolio): Generates the allocation of stocks to buy or sell based on the predicted price changes.
    """    
    def __init__(self, args: StrategyConfig):
        """
        Initialize the Strategy object with the given arguments.

        Args:
            args (StrategyConfig): A configuration object containing necessary parameters.

        Returns:
            None

        Raises:
            None
        """        
        self.args = args
        # train needs around 2-5 years of data
        self.train_past_date = find_business_day(self.args.past_date, -self.args.train_days)
        self.stock_data = {}
        for stock in self.args.stocks: # get historical data for each stock
            data = get_yf_data_some(self.args.stocks, self.train_past_date, self.args.past_date)
            stock_prices = data['Close']
            stock_prices.index = pd.to_datetime(stock_prices.index)  # Ensure datetime format
            stock_prices = stock_prices.asfreq('B').fillna(method='ffill')
            self.stock_data[stock] = stock_prices

    def predict(self, present_date, future_days, stock):
        """
        Predict future stock prices using ARIMA model.

        Args:
            present_date (str): Date for which prediction is being made.
            future_days (int): Number of days into the future to predict.
            stock (str): Stock symbol for which prediction is being made.

        Returns:
            tuple: A tuple containing the log dictionary, percentage change in price, and predicted price.
        """        
        past_date = find_business_day(present_date, -self.args.inference_days)
        data = get_yf_data_some(stock, past_date, present_date)
        stock_prices = data['Close']
        stock_prices.index = pd.to_datetime(stock_prices.index)  # Ensure datetime format
        stock_prices = stock_prices.asfreq('B').fillna(method='ffill')

        best_arima = auto_arima(stock_prices, seasonal=False, trace=True, suppress_warnings=True) # Fit best ARIMA model
        arima_model = ARIMA(stock_prices, order=best_arima.order)
        arima_fit = arima_model.fit()
        price = arima_fit.forecast(steps=future_days)
        pred_pct = (price.iloc[-1] - stock_prices.iloc[-1]) / stock_prices.iloc[-1]
        log = {}
        return log, pred_pct[-1], price[-1]

    def generate_allocation(self, present_date, portfolio):
        """
        Generate allocation for a portfolio based on future price predictions.

        Args:
            date (str): The date for which allocation is generated.
            portfolio (Portfolio): The portfolio object containing information about stocks.

        Returns:
            dict: A dictionary mapping stocks to their corresponding allocation values.
        """        
        future_price, allocation = {}, {}
        for stock in portfolio.tickers:
            _, pred_pct, _ = self.predict(present_date, self.future_days, stock)
            future_price[stock] = pred_pct
        # buy top5 and sell bottom5
        top5 = sorted(future_price.items(), key=lambda x: x[1], reverse=True)[:5]
        bot5 = sorted(future_price.items(), key=lambda x: x[1], reverse=False)[:5]
        buy_qnt = 5; sell_qnt = 5
        for stock, pred in top5:
            if pred > 0:
                allocation[stock] = buy_qnt
                buy_qnt -= 1
        for stock, pred in bot5:
            if pred < 0:
                allocation[stock] = -sell_qnt
                sell_qnt -= 1
        return allocation
    

if __name__ == "__main__":
    args = StrategyConfig(
        inference_days=20,
        future_days=10,
        train_days=365*2,
        past_date="2024-10-10",
        present_date="2024-11-21",
        stocks=['AAPL']
    )
    strategy = ARIMAStrategy(args)
    a, pred, price = strategy.predict(args.present_date, 10, "AAPL")
    print(pred, price)




import os
import importlib

__API_REGISTRY__ = {}

def register_api(name:str):
    def register_model_fn(fn):
        if name in __API_REGISTRY__:
            return __API_REGISTRY__[name]
        __API_REGISTRY__[name] = fn
        return fn
    return register_model_fn




from pypfopt import risk_models, expected_returns
from pypfopt.discrete_allocation import DiscreteAllocation, get_latest_prices
from pypfopt.efficient_frontier import EfficientFrontier
from atradebot.strategies.strategy import Strategy, StrategyConfig
from atradebot.strategies import register_strategy
from atradebot.utils.utils import find_business_day
from atradebot.data.yf_data import get_yf_data_some
from atradebot.params import DATE_FORMAT
from datetime import datetime
import random
# Define a simple strategy that allocates 50% of the portfolio on the first day
def max_sharpe_allocation(data, amount_invest):
    """
    Calculate the maximum Sharpe ratio allocation for a given amount to invest.

    Args:
        data (DataFrame): Historical price data of assets.
        amount_invest (float): Total amount to invest.

    Returns:
        tuple: Allocation dictionary of assets and leftover cash.

    Raises:
        None
    """    
    mu = expected_returns.mean_historical_return(data)
    S = risk_models.sample_cov(data)
    ef = EfficientFrontier(mu, S)
    raw_weights = ef.max_sharpe(risk_free_rate=-10)
    cleaned_weights = ef.clean_weights()
    latest_prices = get_latest_prices(data)
    da = DiscreteAllocation(cleaned_weights, latest_prices, total_portfolio_value=amount_invest)
    allocation, leftover = da.greedy_portfolio()
    return allocation, leftover

@register_strategy("SimpleStrategy")
class SimpleStrategy(Strategy):
    """
    A class representing a simple strategy for trading stocks.

    Attributes:
        args (StrategyConfig): The strategy configuration object containing necessary parameters for the strategy.
        stocks (dict): A dictionary with stock symbols as keys and the number of shares as values.
        cash (list): A list of cash amounts allocated for trading.
        cash_idx (int): An index indicating the current cash amount being used.
        days_interval (int): The number of days interval for cash allocation.
        prev_date (datetime.date): The previous date considered for allocation.
        data (pandas.Series): The stock price data for trading.
    """    
    def __init__(self, args:StrategyConfig):
        """
        Create simple strategy based on sharpe allocation

        Parameters:
            - past_date (datetime.date): date start to backtest in format yyyy-mm-dd
            - present_date (datetime.date): date end to backtest in format yyyy-mm-dd
            - cash (int, optional): amount to invest, defaults to 10000
        """      
        self.args = args          
        cash=args.cash
        self.cash = [cash*0.5, cash*0.3, cash*0.1, cash*0.1] #amount to invest in each rebalance
        self.cash_idx = 0
        delta = self.args.present_date - self.args.past_date
        self.days_interval = delta.days/len(self.cash) #rebalance 4 times
        self.prev_date = self.args.past_date  #previous date for rebalance
    
    def predict(self, present_date, future_days, stock):
        """
        Predict the future stock price based on historical data.

        Args:
            present_date (datetime): The present date for prediction.
            future_days (int): Number of days into the future for prediction.
            stock (str): Stock symbol for prediction.

        Returns:
            tuple: A tuple containing the log dictionary, prediction value, and predicted stock price.

        Note:
            This function relies on external functions find_business_day and get_yf_data_some for data retrieval.
        """        
        past_date = find_business_day(present_date, -self.args.inference_days)
        stock_data = get_yf_data_some(stock, past_date, present_date)
        pred = random.uniform(-1, 1)
        price_ = stock_data['Close'].iloc[-1] * pred
        log = {}
        return log, pred, price_

    def generate_allocation(self, date, portfolio):
        """
        Generate allocation based on the date.
        Args:
            date (pandas.Timestamp): Date to generate allocation.
            portfolio (pandas.DataFrame): Current portfolio (check backtest.py for format).
        Returns:
            dict: Allocation for each stock.
        """        
        delta = date.date() - self.prev_date
        data = portfolio.price_data['Close']
        idx = data.index.get_loc(str(date.date()))
        if date.date() == self.prev_date: #first day
            allocation, leftover = max_sharpe_allocation(data[0:idx], amount_invest=self.cash[self.cash_idx])
            self.cash_idx += 1                        
            return allocation
        elif delta.days > self.days_interval: #rebalance 
            allocation, leftover = max_sharpe_allocation(data[0:idx], amount_invest=self.cash[self.cash_idx])
            self.prev_date = date.date()
            self.cash_idx += 1
            return allocation
        else:
            return {}


if __name__ == "__main__":
    args = StrategyConfig()
    f = SimpleStrategy(args)
    present_date = datetime.strptime("2024-12-10", DATE_FORMAT)
    f.predict(present_date, 10, "AAPL")



# util and helper functions for atradebot

import pandas as pd
from datetime import datetime, timedelta
import yaml
import yfinance as yf
import re
import random
import numpy as np
import json
from atradebot.params import DATE_FORMAT, ALLTICKS, logging, OUTPUT_D
import matplotlib.pyplot as plt
import os

# Print-list of news/LLM answers
def gen_report(backtester, result:pd.DataFrame, fname:str, report:str=""):
    """
    Generate a report by saving a DataFrame to a CSV file.

    Args:
        backtester: The backtester object containing portfolio information.
        result (pd.DataFrame): The DataFrame to be saved as a report.
            contains: 'Date', 'FutureDate', 'Stock', 'Predictions', 'FuturePrice', 'Log': {"prompt", "answer", "news"}
        fname (str): The file name for the CSV file.
        report (str, optional): Additional information to be saved in a text file. Default is an empty string.

    Returns:
        None

    Output:
        Saves the result DataFrame to a CSV file with a specific naming convention.

    Example:
        gen_report(backtester, result, 'report')

    Logging:
        Logs the backtester.portfolio and result information using the logging module.
    """    
    logging.info(backtester.portfolio)
    # Predictions of stocks (rank)
    logging.info(result)
    result.to_csv(f'{OUTPUT_D}/{fname}_data_{str(backtester.args.present_date)}_{backtester.get_strategyName()}.csv', index=False)
    if report:
        with open(f'{OUTPUT_D}/{fname}_report_{str(backtester.args.present_date)}.txt', 'w') as f:
            f.write(report)

def extract_date(filename):
    match = re.search(r'\d{4}-\d{2}-\d{2}', filename)
    if match:
        return datetime.strptime(match.group(), '%Y-%m-%d')
    return None

def get_latest_report(dir):
    paths = os.listdir(dir)
    latest_date = None
    latest_files = []
    for path in paths:
        file_date = extract_date(path)
        if file_date:
            if latest_date is None or file_date > latest_date:
                latest_date = file_date
                latest_files = [path]
            elif file_date == latest_date:
                latest_files.append(path)
    return latest_files 

def get_braces(text):
    """
    Extracts the content inside the first pair of braces in a string.

    Args:
        text (str): The input string from which to extract the content inside the braces.

    Returns:
        str: The content inside the first pair of braces found in the input string, including the braces.
    """    
    match = re.search(r'\{(.*)\}', text)  # Match the first occurrence of {...}
    if match:
        return '{'+match.group(1)+'}'  # Extract the content inside the braces
    return None

def extract_last_number(text):
    """
    Extract the last number from a given text string.

    Args:
        text (str): The input text string containing numeric values.

    Returns:
        float: The last number found in the text string, or None if no numbers are found.
    """    
    matches = re.findall(r'-?\d+\.?\d*', text)
    if matches:
        return float(matches[-1])  # Return the last number as a float
    return None


def get_tickers(text: str):
    """
    Retrieve tickers from a given text.

    Args:
        text (str): The input text to search for tickers.

    Returns:
        list: A list of tickers found in the input text.
    """    
    tickers = [t for t in ALLTICKS if is_word_in_text(t, text)]
    return tickers

def load_args(args, path='config.txt'):
    """
    Loads args from previous run.

    Args:
        args (ArgumentParser): Use args = get_args_parser() to get the args.
        path (str, optional): Path to saved config, defaults to 'config.txt'.

    Returns:
        ArgumentParser: Loaded args.
    """    
    args = args.parse_args([])
    with open(path, 'r') as f:
        args.__dict__ = json.load(f)
    return args

def is_business_day(date:datetime.date):
    """
    Check if date is a business day.

    Args:
        date (datetime.date): Date to check.

    Returns:
        bool: If date is a business day.
    """    
    return bool(len(pd.bdate_range(date, date)))

def find_business_day(start_date:datetime.date, num_days:int=0):
    """
    Add or subtract days from start_date and find the next business day.

    Args:
        start_date (datetime.date): Date to start from.
        num_days (int): Number of days to add/subtract.

    Returns:
        datetime.date: Output date.
    """ 
    out_date = start_date + timedelta(days=num_days)
    while not is_business_day(out_date):
        if num_days < 0:
            out_date -= timedelta(days=1)
        else:
            out_date += timedelta(days=1)
        
    return out_date

#add/subtract business days
def business_days(start_date:datetime.date, num_days:int):
    """
    Add or subtract business days from start_date.

    Args:
        start_date (datetime.date): Date to start from.
        num_days (int): Number of days to add/subtract.

    Returns:
        datetime.date: Output date.
    """    
    current_date = start_date
    business_days_added = 0
    while business_days_added < abs(num_days):
        if num_days < 0:
            current_date -= timedelta(days=1)
        else:
            current_date += timedelta(days=1)
        if current_date.weekday() < 5:  # Monday to Friday
            business_days_added += 1
    return current_date


def pd_append(df, dict_d):
    """
    Append dictionary to dataframe.

    Args:
        df (pandas.DataFrame): Pandas dataframe.
        dict_d (dict): Dictionary to append.
    
    Returns:
        pandas.DataFrame: Pandas dataframe with appended dictionary.
    """ 
    return pd.concat([df, pd.DataFrame.from_records([dict_d])])

def is_word_in_text(word, text):
    # Use regex with word boundaries (\b) for full-word matching
    """
    Check if a word is in a given text.

    Args:
        word (str): The word to search for.
        text (str): The text in which to search for the word.

    Returns:
        bool: True if the word is found in the text, False otherwise.
    """    
    pattern = r'\b{}\b'.format(re.escape(word))
    return bool(re.search(pattern, text))

def plot_stock(stocks:dict, show=False):
    # stocks: {"SPY":data_spy/data_spy[0], #normalize gains "ThisPortfolio":data_my/init_capital}
    # Plotting configuration
    """
    Plot stock prices for each stock in the input dictionary.

    Args:
        stocks (dict): A dictionary where keys are stock names and values are pandas Series with date index.
        show (bool, optional): Whether to display the plot. Default is False.

    Returns:
        matplotlib.pyplot: The plot showing stock price comparison.

    Example:
        plot_stock({'AAPL': apple_stock_data, 'GOOGL': google_stock_data}, show=True)

    Note:
        Requires matplotlib library for plotting.
    """    
    plt.figure(figsize=(10, 6))  # Set the figure size
    for k, stock_data in stocks.items():
        plt.plot(stock_data.index, stock_data.values, label=k)
    plt.xlabel('Date')
    plt.ylabel('Price')
    plt.title('Stock Price Comparison')
    plt.legend()
    if show:
        plt.show()
    return plt

def get_config(cfg_file):
    """
    Reads a YAML configuration file and returns its content as a Python dictionary.

    Args:
        cfg_file (str): The path to the YAML configuration file.

    Returns:
        dict: A dictionary containing the configuration parameters.
    """    
    with open(cfg_file, 'r') as ymlfile:
        cfg = yaml.load(ymlfile, Loader=yaml.FullLoader)
    return cfg

def find_peaks_valleys(array):
    """
    Find peaks and valleys.

    Args:
        array (list): List of numbers.

    Returns:
        list: Reduced list of peaks and valleys. List of ids of the array.

    Raises:
        None
    """    
    peaks = []
    valleys = []
    off = 5
    for i in range(off, len(array) - off, off):
        if array[i - off] < array[i] > array[i + off]:
            peaks.append(i)
        elif array[i - off] > array[i] < array[i + off]:
            valleys.append(i)
    return peaks, valleys


def filter_points(data, peak_idx, valley_idx):
    """
    Ignore peaks and valleys that are too close to each other.

    Args:
        data (pandas): Data from yfinance.download.
        peak_idx (List[int]): List of indices for peaks.
        valley_idx (List[int]): List of indices for valleys.

    Returns:
        List[int]: Reduced list of peaks and valleys.
    """       
    len_min = min(len(peak_idx), len(valley_idx))
    idx = 0
    coef_var = np.std(data)/np.mean(data)
    peak_idx_n, valley_idx_n = [], []
    while idx < len_min:
        abs_diff = abs(data[peak_idx[idx]]-data[valley_idx[idx]])
        min_diff = min(data[peak_idx[idx]], data[valley_idx[idx]])
        percent = abs_diff/min_diff
        if percent > coef_var*0.2: 
            peak_idx_n.append(peak_idx[idx])
            valley_idx_n.append(valley_idx[idx])
        idx += 1
    return peak_idx_n, valley_idx_n
    
# run financial sentiment analysis model
def get_forecast(stock, date, add_days=[21, 5*21, 12*21]):
    """
    Get forecast for stock on date.

    Args:
        stock (str): Stock id.
        date (datetime): Date in datetime format.
        add_days (list, optional): Number of days to add for forecasting in a list, defaults to [21, 5*21, 12*21].

    Returns:
        List[int]: Forecast list based on add_days. E.g., add_days=[21, 5*21, 12*21] return is stock gain/loss in [1mon, 5mon, 1 yr]. Higher than 1 indicates percent increase, lower than 1 indicates percent decrease.
    """
    s = date
    e = business_days(s, +3)
    data = yf.Ticker(stock)
    hdata = data.history(start=s.strftime("%Y-%m-%d"),  end=e.strftime("%Y-%m-%d"))
    price = hdata['Close'].mean()

    forecast = [0, 0, 0]
    for idx, adays in enumerate(add_days): #add business days
        s = business_days(s, adays)#look into future
        e = business_days(s, +3)
        hdata = data.history(start=s.strftime("%Y-%m-%d"),  end=e.strftime("%Y-%m-%d"))
        forecast[idx] = hdata['Close'].mean()/price
    
    return forecast


def get_mentionedtext(keyword, text, context_length=512):
    """
    Get text around keyword given the number of characters to extract 
    before and after the keyword.

    Args:
        keyword (str): Keyword to search for in text.
        text (str): Text to search for keyword.
        context_length (int, optional): Number of words before and after the keyword, defaults to 512.

    Returns:
        str: Collected text around keyword.
    """    
    delimiter = '.'
    # Create a regex pattern to match the keyword with surrounding text
    # pattern = r"(.{0,%d})(%s)(.{0,%d})" % (context_length, re.escape(keyword), context_length)
    pattern = r"([^%s]+%s[^%s]+)%s" % (delimiter, re.escape(keyword), delimiter, delimiter)

    # Find all matches in the text
    matches = re.findall(pattern, text)
    f_text = ''
    for match in matches:
        match = match.replace("\n", "")
        f_text += match
        if len(f_text) > context_length:
            break
    return f_text



def gen_rand_alloc(n_stock=5):
# Generate random percentages that sum up to 100
    """
    Generate random allocation percentages for a list of stocks.

    Args:
        n_stock (int): The number of stocks for which to generate allocation percentages. Default is 5.

    Returns:
        list: A list of random allocation percentages for each stock, summing up to 100.

    Note:
        The last stock's allocation percentage is calculated to ensure that the total sum is 100.
    """    
    percentages = [random.randint(1, 100) for _ in range(n_stock - 1)]
    remaining_percentage = 100 - sum(percentages)
    percentages.append(remaining_percentage)
    return percentages


def get_datetime(date_str):
    """
    Convert a date string to a datetime object.

    Args:
        date_str (str or datetime): A date string in the format specified by DATE_FORMAT or a datetime object.

    Returns:
        datetime.date: The date extracted from the input date string.
    """    
    if isinstance(date_str, str):
        return datetime.strptime(date_str, DATE_FORMAT).date()
    return date_str


from abc import ABC, abstractmethod
import datetime
from dataclasses import dataclass, asdict, fields
import pandas

@dataclass
class StrategyConfig:
    # arguments for strategy
    """
    A dataclass representing the configuration for a trading strategy.

    Attributes:
        data (pandas.DataFrame): The historical data for trading.
        past_date (datetime.date): The starting date for historical data.
        present_date (datetime.date): The current date for trading.
        stocks (list): The list of stocks to trade.
        inference_days (int): Number of days to consider for inference.
        future_days (int): Number of days to predict into the future.
        train_days (int): Number of days to train the model. history-data used in arima and strategies that train as it goes [Rolling Forecast]
        cash (int): Initial amount of capital for trading.
        newsapi (str): API to use for news data.
        strategy (str): The trading strategy to use.
    """    
    past_date: datetime.date = datetime.date(2023, 5, 1)
    present_date: datetime.date = datetime.date(2023, 12, 10)
    inference_days: int = 10
    future_days: int = 10
    train_days: int = 365*3  
    newsapi: str = 'dataset'
    strategy: str = 'SimpleStrategy'
    run_trade: bool = False
    cash: int = 10000
    
    

class Strategy(ABC):
    
    @abstractmethod   
    def predict(self, present_date:datetime.date, future_days:int, stock:str):
        # returns: model output, percentage change, predicted price
        """
        Predict the stock price for a given future date.

        Args:
            present_date (datetime.date): The present date for which the prediction is made.
            future_days (int): The number of days in the future for which the prediction is made.
            stock (str): The stock symbol for which the prediction is made.

        Raises:
            NotImplementedError: This is an abstract method that must be implemented in a subclass.
        """        
        pass # Predict function not implemented

    @abstractmethod
    def generate_allocation(self, date:datetime.date, portfolio):
        # returns: allocation: dict{"stock": number of stocks to buy/sell }
        """
        Generate allocation for a specific date and portfolio.

        Args:
            date (datetime.date): The date for which allocation needs to be generated.
            portfolio (object): The portfolio for which allocation needs to be generated.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """        
        pass # Generate allocation function not implemented


# regression tests

import pytest
from atradebot.data.yf_data import get_yf_data_all, get_yf_data_some
from atradebot.data.get_news import get_news
from atradebot.utils.portfolio import Portfolio
from atradebot.utils.utils import get_datetime, find_business_day
from atradebot.utils.backtest import backtest
from datetime import datetime, date
from atradebot.params import DATE_FORMAT
from atradebot.utils.search import stock_research, findstocks
from atradebot.strategies.strategy import StrategyConfig
from atradebot.main import manage
from atradebot.params import ROOT_D

@pytest.mark.test
def test_portfolio():
    portfolio_path = f'{ROOT_D}/sd/test_portfolio.csv'
    args = StrategyConfig(
        past_date='2023-08-01',
        present_date='2023-08-22',
        inference_days=5,
        future_days=5,
        strategy='SimpleStrategy'
    )
    portfolio = Portfolio(portfolio_path, args)
    beta, alpha = portfolio.calc_beta()
    print("The portfolio beta is", round(beta, 4))
    print("The portfolio alpha is", round(alpha,5))

    bdays = 3
    sloss = 0.4
    print(f"The stocks going down past {bdays} days: ", portfolio.stop_down(bdays))
    print(f"The stocks went down {sloss} from cost basis are: ", portfolio.stop_loss(loss=sloss))

    portfolio.update_portfolio(day=get_datetime('2023-08-08'))
    print(portfolio)
    portfolio.buy_stock('AAPL', 10)
    print(portfolio)
    portfolio.sell_stock('AAPL', 5)
    print(portfolio)
    #buy new stock
    portfolio.buy_stock('TSLA', 10)
    print(portfolio)
    portfolio.sell_stock('TSLA', 5)
    print(portfolio)

@pytest.mark.test
def test_get_yfdata():
    data = get_yf_data_some("AAPL", "2024-10-15", "2024-10-17")
    assert data is not None

@pytest.mark.test
def test_search():
    stocks, answ = stock_research()

@pytest.mark.test
def test_getnews():
    news = get_news(stock='A', num_results=10, time_period=[], news_source='dataset')

# main apps
@pytest.mark.apps
def test_findstocks():
    inference_days = 5
    present_date = date.today()
    past_date = find_business_day(present_date, -inference_days)
    args = StrategyConfig(
        past_date=past_date,
        present_date=present_date,
        inference_days=inference_days,
        future_days=10, 
        strategy='FinLlama3Strategy'
    )
    findstocks(args)

@pytest.mark.apps
def test_managestocks():
    inference_days = 5
    present_date = date.today()
    past_date = find_business_day(present_date, -inference_days)
    args = StrategyConfig(
        past_date=past_date,
        present_date=present_date,
        inference_days=inference_days,
        future_days=10, 
        strategy='FinLlama3Strategy'
    )
    manage(args, portfolio_path=f'{ROOT_D}/sd/test_portfolio.csv')
    

TEST_PAST_DATE = datetime.strptime('2023-08-01', DATE_FORMAT).date()
TEST_PRESENT_DATE = datetime.strptime('2023-08-22', DATE_FORMAT).date() 
# model tests
@pytest.mark.backtest
def test_backtest_alloc_simple():
    past_date = TEST_PAST_DATE
    present_date = TEST_PRESENT_DATE
    args = StrategyConfig(
        past_date=past_date,
        present_date=present_date,
        inference_days=10,
        future_days=10, 
        strategy='SimpleStrategy'
    )
    backtest(
        portfolio_path=f'{ROOT_D}/sd/test_portfolio.csv',
        args=args,
        eval_mode='alloc',
        plot=False)

@pytest.mark.backtest
def test_backtest_predict_simple():
    past_date = TEST_PAST_DATE
    present_date = TEST_PRESENT_DATE
    args = StrategyConfig(
        past_date=past_date,
        present_date=present_date,
        inference_days=5,
        future_days=5, 
        strategy='SimpleStrategy'
    )
    backtest(
        portfolio_path=f'{ROOT_D}/sd/test_portfolio.csv',
        args=args,
        eval_mode='predict',
        plot=False)

    
    
  



# based on: https://github.com/AI4Finance-Foundation/FinGPT/tree/master/fingpt/FinGPT_Forecaster
# Huggingface Llama3 based serving


import torch
from peft import PeftModel
from collections import defaultdict
from datetime import date, datetime, timedelta
from transformers import AutoTokenizer, AutoModelForCausalLM, TextStreamer
from atradebot.data.yf_data import get_yf_data_some
from atradebot.params import DATE_FORMAT
from atradebot.strategies import register_strategy
from atradebot.strategies.strategy import Strategy
from atradebot.utils.utils import find_business_day, extract_last_number
from atradebot.data.llm_prompt import construct_prompt, format_chat_template
from atradebot.trainer.train_llama3 import BASEMODEL, LORAMODEL, RESPONSE_TOKENS
from atradebot.strategies.strategy import StrategyConfig

@register_strategy("FinLlama3Strategy")
class FinLlama3Strategy(Strategy): 
    """
    A class representing a financial trading strategy using language model inference.

    Attributes:
        newsapi (NewsAPI): The NewsAPI instance used to retrieve news for the strategy.
        inference_days (int): The number of days of historical data to use for model inference.
        future_days (int): The number of days into the future to predict stock prices.
        model (AutoModelForCausalLM): The pretrained language model for generating predictions.
        tokenizer (AutoTokenizer): The tokenizer for processing input text.
        streamer (TextStreamer): The TextStreamer instance for handling text data streams.
    """    
    def __init__(self, args:StrategyConfig):
        """
        Initialize the Strategy object with the given arguments.

        Args:
            args (StrategyConfig): Object containing configuration parameters for the strategy.

        Returns:
            None

        Raises:
            None
        """        
        self.newsapi = args.newsapi
        self.inference_days = args.inference_days
        self.future_days = args.future_days
        self.model = AutoModelForCausalLM.from_pretrained(
            BASEMODEL,
            trust_remote_code=True, 
            device_map="auto",
            torch_dtype=torch.float16)
        # self.model = PeftModel.from_pretrained(self.bmodel,LORAMODEL)
        self.model = self.model.eval()
        self.tokenizer = AutoTokenizer.from_pretrained(BASEMODEL)
        self.streamer = TextStreamer(self.tokenizer)
    
    def generate_allocation(self, present_date:datetime.date, portfolio):
        """
        Generate allocation for a portfolio based on predicted future prices.

        Args:
            date (datetime.date): The date for which the allocation is generated.
            portfolio (object): An object representing the portfolio containing stock tickers.

        Returns:
            dict: A dictionary containing the allocation for each stock ticker.

        Raises:
            None
        """        
        future_price, allocation = {}, {}
        for stock in portfolio.tickers:
            _, pred_pct, _ = self.predict(present_date, self.future_days, stock)
            future_price[stock] = pred_pct
        # buy top5 and sell bottom5
        top5 = sorted(future_price.items(), key=lambda x: x[1], reverse=True)[:5]
        bot5 = sorted(future_price.items(), key=lambda x: x[1], reverse=False)[:5]
        buy_qnt = 5; sell_qnt = 5
        for stock, pred in top5:
            if pred > 0:
                allocation[stock] = buy_qnt
                buy_qnt -= 1
        for stock, pred in bot5:
            if pred < 0:
                allocation[stock] = -sell_qnt
                sell_qnt -= 1
        return allocation

    def predict(self, present_date:datetime.date, future_days:int, stock:str): 
        # return: log: dict with model input and output to save, pred: prediction, price_: predicted price
        """
        Predict the future stock price using a language model.

        Args:
            present_date (datetime.date): The present date for which the prediction is being made.
            future_days (int): Number of future days for which the prediction is needed.
            stock (str): Stock symbol for which the prediction is being made.

        Returns:
            tuple: A tuple containing log information, prediction percentage, and the predicted price.

        Raises:
            None
        """        
        past_date = find_business_day(present_date, -self.inference_days)
        stock_data = get_yf_data_some(stock, past_date, present_date)
        prompt, news = construct_prompt(stock, present_date, past_date, futuredays=future_days, newsapi=self.newsapi)
        row = {"instruction": prompt}
        prompt = format_chat_template(self.tokenizer, row)

        inputs = self.tokenizer(prompt["text"], return_tensors='pt', padding=False)
        inputs = {key: value.to(self.model.device) for key, value in inputs.items()}            
        res = self.model.generate(
            **inputs, max_length=128_000, do_sample=True,
            eos_token_id=self.tokenizer.eos_token_id,
            use_cache=True, 
            # streamer=self.streamer
        )
        output = self.tokenizer.decode(res[0])
        ii = output.find(RESPONSE_TOKENS)
        output = output[ii:]
        print('OUTPUT: ', output)
        answer = output.split('%')[0]
        pred = extract_last_number(answer)
        torch.cuda.empty_cache()
        cur_price = stock_data['Close'].iloc[-1][-1]
        if pred is None:
            print("Failed to get prediction from the model.")
            return None, 0, cur_price
        
        price_ = cur_price*(pred/100) + cur_price
        log = {"prompt": prompt, "answer": answer, "news": news}
        return log, pred, price_        
        
        
# example answer:
if __name__ == "__main__":
    args = StrategyConfig()
    f = FinLlama3Strategy(args)
    present_date = datetime.strptime("2024-12-10", DATE_FORMAT)
    a, pred, price = f.predict(present_date, 10, "AAPL")
    print(pred, price)


# add different source to gather news and data here
# function input: stock, date, num_results
# function output: list of dict {"link", "title", "snippet", "date", "source", "text", "stock"}

import pandas as pd
# import tweepy
from argparse import ArgumentParser
import os
from datetime import date, datetime, timedelta
import requests
from bs4 import BeautifulSoup
import trafilatura
from trafilatura.settings import use_config
import finnhub
from tenacity import retry, stop_after_attempt, wait_fixed
from atradebot.data.__init__ import register_api
import traceback
from datasets import load_dataset
from gpt_researcher import GPTResearcher
import asyncio
from atradebot.params import NEWSCACHE
from atradebot.params import DATE_FORMAT, FHUB_API, HF_DATASET


LOADED_DATASET = load_dataset(HF_DATASET, split='train')
# LOADED_DATASET = load_from_disk(HF_DATASET)
LOADED_DATASETPD = LOADED_DATASET.to_pandas()

trafilatura_config = use_config()
trafilatura_config.set("DEFAULT", "EXTRACTION_TIMEOUT", "0")
finnhub_client = finnhub.Client(api_key=FHUB_API)


# create research news document for stocks
async def run_gptresearch(prompt):
    # Report Type
    """
    Run a GPTResearch operation using a specified prompt.

    Args:
        prompt (str): The prompt to use for the GPTResearch operation.

    Returns:
        str: The generated research report.

    Raises:
        None
    """    
    report_type = "research_report"

    # Initialize the researcher
    researcher = GPTResearcher(query=prompt, report_type=report_type, config_path=None)
    # Conduct research on the given query
    await researcher.conduct_research()
    # Write the report
    report = await researcher.write_report()
    
    return report


@register_api('google')
@retry(
    stop=stop_after_attempt(2),   # Retry up to 2 times
    wait=wait_fixed(60)  # Wait 60 seconds between retries
)
@NEWSCACHE.memoize(expire=604800)
def get_google_news(stock, num_results=10, time_period=[]):
    """
    Collect google search text and limit to 100 results per day.

    Args:
        stock (List[str]): List of stock ids
        num_results (int, optional): Number of news to collect, defaults to 10
        time_period (list, optional): Time period=['2019-06-28' (start_time), '2019-06-29' (end_time)], defaults to []

    Returns:
        list of news dict {"link", "title", "snippet", "date", "source", "text", "stock"}, search query, soup html searched
        str, html
    """     
    query = stock
    headers = {
            "User-Agent":
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.54 Safari/537.36"
        }
    if time_period:
        query += f"+before%3A{time_period[1]}+after%3A{time_period[0]}" # add time range     
    search_req = "https://www.google.com/search?q="+query+"&gl=us&tbm=nws&num="+str(num_results)+""
    #https://developers.google.com/custom-search/docs/xml_results#WebSearch_Request_Format
    news_results = []
    # get webpage content
    response = requests.get(search_req, headers=headers)
    soup = BeautifulSoup(response.content, "html.parser")
    for el in soup.select("div.SoaBEf"):
        sublink = el.find("a")["href"]
        downloaded = trafilatura.fetch_url(sublink)
        html_text = trafilatura.extract(downloaded, config=trafilatura_config)
        if html_text:
            date_ = el.select_one(".LfVVr").get_text()
            if 'ago' in date_:
                continue 
            date_i = datetime.strptime(date_, "%b %d, %Y")
            date_i = date_i.strftime(DATE_FORMAT)
            news_results.append(
                {
                    "link": el.find("a")["href"],
                    "title": el.select_one("div.MBeuO").get_text(),
                    "snippet": el.select_one(".GI74Re").get_text(),
                    "Date": date_i,
                    "source": el.select_one(".NUnG9d span").get_text(),
                    "News": html_text,
                    "Tickers": stock,
                }
            )
    return news_results, search_req, soup

@register_api('finhub')
@retry(stop=stop_after_attempt(2), wait=wait_fixed(60))
@NEWSCACHE.memoize(expire=604800)
def get_finhub_news(stock, num_results=10, time_period=[]):
    """
    collect from finhub news: https://finnhub.io/
    only up to 1 year old news and 60 requests per minute

    Args:
        stock (List[str]): list of stock ids
        num_results (int, optional): number of news to collect, defaults to 10
        time_period (list, optional): time_period=['2019-06-28' (start_time), '2019-06-29' (end_time)], defaults to []

    Returns:
        list of dict, None, None: list of news in a dict{"link", "title", "snippet", "date", "source", "text", "stock"}

    Raises:
        None
    """    
    if len(time_period) == 0:
        time_period = [str(date.today() - timedelta(days=10)), str(date.today())]

    results = finnhub_client.company_news(stock, _from=time_period[0], to=time_period[1])
    news_results = []
    for result in results:
        # try:
        #     downloaded = trafilatura.fetch_url(result['url'])
        # except:
        #     continue
        # html_text = trafilatura.extract(downloaded, config=trafilatura_config)
        html_text = result['summary']
        news_results.append(
            {
                "link": result['url'],
                "title": result['headline'],
                "snippet": result['summary'],
                "Date": datetime.fromtimestamp(result['datetime']).strftime(DATE_FORMAT),
                "source": result['source'],
                "News": html_text,
                "Tickers": stock,
            }
        )
    search_req, soup = None, None
    return news_results, search_req, soup


@register_api('dataset')
def get_dataset_news(stock, num_results=10, time_period=[]):
    """
    Get news dataset for a specific stock.

    Args:
        stock (str): The stock symbol for which to retrieve news data.
        num_results (int): The number of news results to return. Default is 10.
        time_period (list): A list containing a start and end date for filtering news data. Default is an empty list.

    Returns:
        tuple: A tuple containing news results as a list of dictionaries, search request, and soup.
    """    
    dataset = LOADED_DATASETPD
    dataset = dataset[dataset['Tickers'] == stock]
    if len(dataset) == 0:
        return [], None, None
    
    if len(time_period) == 0:
        dataset = dataset.sample(n=num_results)
    else:
        dataset = dataset[(dataset['Date'] >= time_period[0]) & (dataset['Date'] <= time_period[1])]

    news_results = dataset.to_dict(orient='records')
    search_req, soup = None, None
    return news_results, search_req, soup

@register_api('search')
@NEWSCACHE.memoize(expire=604800)  
def search_news(stock:str, num_results=10, time_period=[]):
    # get news for a stock (inference+dataset)
    """
    Search for news related to a specific stock within a time period.

    Args:
        stock (str): The stock symbol to search for news.
        num_results (int): Number of news results to retrieve. Default is 10.
        time_period (list): List containing two elements - the start and end date of the time period.

    Returns:
        list: A list of dictionaries containing information about the news articles found.

    Raises:
        None
    """    
    sysPrompt = 'You are a financial advisor. For the period between {} and {}, search the web for top {} \
        news that can alter the stock price of {}. ' 
    query = sysPrompt.format(num_results, time_period[0], time_period[1], stock)
    answer = asyncio.run(run_gptresearch(query))
    result = [{"link": '',
                    "title": '',
                    "snippet": '',
                    "Date": time_period[1],
                    "source": 'gptresearch',
                    "News": answer,
                    "Tickers": stock}]
    return result


def get_news(stock, time_period:list, num_results:int=10, news_source:str='google'):
    """
    Wrapper function to get news from different sources.

    Args:
        stock (List[str]): List of stock ids.
        time_period (list, optional): Time period=['2019-06-28' (start_time/past), '2019-06-29' (end_time/present)]. Defaults to [].
        num_results (int, optional): Number of news to collect. Defaults to 10.
        news_source (str, optional): News source ['google', 'finhub', 'newsapi', 'search']. Defaults to 'finhub'.

    Returns:
        list of news dict: {"link", "title", "snippet", "date", "source", "text", "stock"}, search query, soup html searched.
        str: Search query.
        html: Soup html searched.
    """    
    try:
        news = []
        if news_source == 'all':
            news, _, _ = get_google_news(stock=stock, num_results=num_results, time_period=time_period)
            news1, _, _ = get_finhub_news(stock=stock, num_results=num_results, time_period=time_period)
            news += news1
        elif news_source == 'google':
            news, _, _ = get_google_news(stock=stock, num_results=num_results, time_period=time_period)
        elif news_source == 'finhub':
            news, _, _ = get_finhub_news(stock=stock, num_results=num_results, time_period=time_period)
        elif news_source == 'dataset': #use news from dataset
            news, _, _ = get_dataset_news(stock=stock, num_results=num_results, time_period=time_period)
        else: #news_source == 'search': # generated from search.py
           news, _, _ = search_news(stock=stock, num_results=num_results, time_period=time_period)

        if news == []:
            print(f"Can't collect news for {stock}")
    except:
        traceback.print_exc()
        return []

    return news

if __name__ == "__main__":
    news = get_news(stock='AAPL', time_period=['2020-01-01', '2020-01-02'], num_results=3, news_source='google')
    print(news)



# Backtesting module for benchmarking portfolio strategiesallocation 

import pandas as pd
import os
from argparse import ArgumentParser
import matplotlib.pyplot as plt
from datetime import datetime
from atradebot.params import DATE_FORMAT, ROOT_D
from atradebot.utils.portfolio import Portfolio, test_portfolio
from atradebot.strategies import get_strategy
from atradebot.utils.utils import find_business_day, get_datetime, plot_stock
from atradebot.strategies import __MODEL_REGISTRY__
from sklearn.metrics import mean_squared_error,mean_absolute_error,r2_score
from atradebot.strategies.strategy import StrategyConfig


pd.options.mode.chained_assignment = None

class Backtester():
    """
    A class representing a backtester for evaluating portfolio strategies.

    Attributes:
        portfolio_path (str): Path to the portfolio csv file.
        args (StrategyConfig): Strategy configuration for the backtester.
        zero (bool): Flag indicating whether to zero allocation.
        outdir (str): Output directory for the backtester results.
    """    
    def __init__(self, 
                 portfolio_path:str,
                 args:StrategyConfig,
                 zero:bool=False,
                 outdir:str='./'
                 ):
        """
        Create backtester to evaluate portfolio strategies.

        Args:
            portfolio_path (str): Path to the portfolio csv file.
            strategy (str): Strategy to use.
            past_date (datetime.date): Start of evaluation period (past).
            present_date (datetime.date): End of evaluation period.
            inference_days (datetime.date): Number of days to look back for model inference.
            outdir (str): Output directory.
        """    
        self.args = args    
        past_date = get_datetime(args.past_date)
        present_date = get_datetime(args.present_date)
        self.portfolio = Portfolio(portfolio_path, 
                                    args, 
                                    outdir=outdir)
        # for backtesting start with zero allocation to test if strategy can generate allocation and predictions.
        # Used for backtest(), findstocks(), where we start with empty portfolio
        # Only in manage() we start with a populated portfolio
        if zero:
            self.portfolio.zero_allocation()

        delta = present_date - past_date
        self.future_days = args.future_days
        self.inference_days = abs(args.inference_days) # number of days to look back for model inference
        assert delta.days >= self.inference_days, "inference_days should be less than the evaluation period"
        self.pastday = find_business_day(past_date, self.inference_days) # start of inference period
        self.strategy = get_strategy(args.strategy)(args)
        self.price_data = self.portfolio.price_data # get stock price data
        self.tickers = self.portfolio.tickers  
        self.activity = [] #list of (date, allocation)
        self.portfolio_perf = [] # total portfolio value over time
    
    def get_strategyName(self):
        """
        Get the name of the strategy class.

        Returns:
            str: The name of the strategy class.
        """        
        return self.strategy.__class__.__name__
    
    def rank_predict(self, date, future_days:list):
        """
        Predict future stock price ranks for each ticker.

        Args:
            date (str): The date for which the predictions are made.
            future_days (list): A list of future days to predict for.

        Returns:
            pandas.DataFrame: A DataFrame containing predictions for each ticker on each future day.
            contains: 'Date', 'FutureDate', 'Stock', 'Predictions', 'FuturePrice', 'Log': {"prompt", "answer", "news"}
        """        
        results = []
        for ticker in self.tickers: # for each stock
            for fday in future_days: # for eval dates
                log, pred, price = self.strategy.predict(date, fday, ticker)
                results.append({'Date':date, 'FutureDate':fday, 'Stock':ticker, 'Predictions':pred, 'FuturePrice':price, 'Log':log})
        df = pd.DataFrame(results)
        df.sort_values(by='Predictions', ascending=False, inplace=True)
        return df
        
    def eval_predict(self, plot=False):
    # measure how well model predict price change for stocks
        """
        Evaluate and predict the stock prices using a given strategy.

        Args:
            plot (bool, optional): Whether to plot and save the predictions. Default is False.

        Returns:
            list: A list of dictionaries containing evaluation metrics and predictions for each stock.

        Raises:
            None
        """        
        idx = self.price_data.index.get_loc(str(self.pastday))
        results = []
        # num_evals = int(len(self.price_data)*0.2)
        num_evals = len(self.price_data)
        for ticker in self.tickers: # for each stock
            tgts, preds = [], []
            for i in range(idx, num_evals - 1): # for eval dates
                date = self.price_data.index[i]
                tgt = self.price_data.iloc[i+1]['Close'][ticker]
                _, pred, price = self.strategy.predict(date, self.future_days, ticker)
                tgts.append(tgt)
                preds.append(price)
            MSE = mean_squared_error(preds, tgts)
            RMSE = MSE**0.5
            MAE = mean_absolute_error(preds, tgts)
            R2 = r2_score(preds, tgts)
            if plot:
                plt.plot(tgts, label=f'Target_{ticker}')
                plt.plot(preds, label=f'Predicted_{ticker}')
                plt.legend()
                plt.savefig(f'{ROOT_D}/sd/predict_{ticker}_{str(self.strategy.__class__.__name__)}.png')
                plt.close()
            results.append({'Stock':ticker, 'MSE': MSE, 'RMSE': RMSE, 'MAE': MAE, 'R2': R2, 'targets':tgts, 'predictions':preds})
    
        return results

    def run_backtest(self):
    # measure how well the strategy performs to trade and allocate stocks 
        # Initialize portfolio with initial capital
        """
        Run backtest on a financial strategy using historical price data.

        Args:
            self: The object instance.
    
        Returns:
            None

        Note:
            This function updates the portfolio based on the generated allocation from the strategy for each date in the historical price data.

        Raises:
            None
        """        
        idx = self.price_data.index.get_loc(str(self.pastday))
        for i in range(idx, len(self.price_data) - 1):
            # Retrieve current date and price
            date = self.price_data.index[i]
            # Call strategy to determine portfolio allocation
            allocation = self.strategy.generate_allocation(date, self.portfolio) # dict{stock: alloc} 
            self.portfolio.update_portfolio(date)
            act = self.portfolio.execute(allocation)
            # Calculate total portfolio value
            if act:
                self.activity.append((date, allocation))
            self.portfolio_perf.append({'Date':date, 'Price': self.portfolio.get_portfolio_value()})

        assert len(self.portfolio_perf) > 0, "run_backtest: no portfolio performance data available"

        self.portfolio_perf = pd.DataFrame(self.portfolio_perf).set_index('Date')

def backtest(
        portfolio_path:str,
        args:StrategyConfig,
        eval_mode:str='predict',
        plot:bool=False,
):
    """
    run backtest

    Args:
        plot (bool): Enable plot to return or not.

    Returns:
        tuple: plot, backtester class, portfolio value, data for each stock.

    Raises:
        None
    """
    # Create a portfolio backtester instance
    init_capital = args.cash
    if not portfolio_path: # TODO: no search new stocks backtesting yet
        stocks=['AAPL', 'NVDA', 'TSLA', 'MSFT', 'GOOGL', 'AMZN']
        portfolio_path = test_portfolio(stocks)

    backtester = Backtester(portfolio_path, args=args, zero=True)
    if eval_mode == 'predict':
        backtester.eval_predict(plot=True)
        return
    
    # Run the backtest using the simple strategy
    backtester.run_backtest()

    # Retrieve the portfolio value
    portfolio_value = backtester.portfolio.get_portfolio_value()
    
    if plot:
        idx = backtester.portfolio.price_data.index.get_loc(str(backtester.pastday))
        data_spy = backtester.portfolio.benchmark_price['Close'][idx:]
        data_my = backtester.portfolio_perf[idx:]
        cost = backtester.portfolio.get_cost()
        qnt = int(cost/data_spy.values[0][0])
        # dict of pandas series: Date: Price
        plt = plot_stock({
                "SPY":data_spy*qnt, #normalize gains
                "ThisPortfolio":data_my}, show=False)
        plt.savefig(f'{ROOT_D}/sd/backtest_{backtester.get_strategyName()}.png')
        print('Activity: ', backtester.activity)
    else:
        plt = None

    print(portfolio_value)
    print(backtester.portfolio)

def get_args(raw_args=None):
    """
    Get the command line arguments for the program.

    Args:
        raw_args (list): A list of raw arguments to parse. Default is None.

    Returns:
        argparse.Namespace: An object containing the parsed arguments.

    Raises:
        None.

    Description:
        This function creates an ArgumentParser object to define and parse command line arguments
        for the program. It sets various optional arguments for the strategy, number of days for inference,
        number of days in the future to predict, start and end dates for trading analysis, evaluation mode,
        and path to the portfolio. It then parses the raw arguments provided and returns an object
        containing the parsed arguments.

        Example Usage:
        raw_args = ['--strategy', 'RandomStrategy', '--inference_days', '3']
        args = get_args(raw_args)
    """    
    parser = ArgumentParser(description="parameters")
    parser.add_argument('--strategy','-s', type=str, default='ManualStrategy', help=f'Modes: {__MODEL_REGISTRY__.keys()}')
    parser.add_argument('--inference_days','-i', type=int, default=1, help='number of days for inference')
    parser.add_argument('--future_days','-f', type=int, default=1, help='number of days in the future to predict')
    parser.add_argument('--past_date', type=str, default="2024-04-01", help='eval start date for trading analysis (past)')
    parser.add_argument('--present_date', type=str, default="2025-01-30", help='eval end data for trading analysis')
    parser.add_argument('--newsapi','-n', type=str, default="dataset", help='newsapi: finhub, newsapi, google')
    parser.add_argument('--eval_mode', type=str, default="allocation", help='eval modes: predict and allocation')
    parser.add_argument('--portfolio_path', type=str, default="", help='path to portfolio')
    args = parser.parse_args(raw_args)
    return args

if __name__ == "__main__":

    args = get_args()
    print(args)
    past_date = datetime.strptime(args.past_date, DATE_FORMAT).date()
    present_date = datetime.strptime(args.present_date, DATE_FORMAT).date()
    config = StrategyConfig(
        past_date=past_date,
        present_date=present_date,
        inference_days=args.inference_days,
        future_days=args.future_days,
        strategy=args.strategy,
        newsapi=args.newsapi,
    )

    backtest(
        portfolio_path=args.portfolio_path,
        args=config,
        eval_mode=args.eval_mode,
        plot=True)
    



import asyncio
from datetime import datetime
from atradebot.utils.utils import get_tickers, gen_report
from atradebot.utils.backtest import Backtester  
from atradebot.data.yf_data import get_price
from atradebot.data.get_news import run_gptresearch
from atradebot.strategies.strategy import StrategyConfig
from argparse import ArgumentParser
from atradebot.strategies import __MODEL_REGISTRY__
from datetime import date, datetime
from atradebot.params import DATE_FORMAT, ROOT_D
from atradebot.utils.utils import find_business_day, gen_report
from atradebot.utils.alpaca_trade import alpaca_execute
import os

MARKET_CAP = 'large or medium cap'
SECTOR = 'information technology'

# findstocks: function to search for stock ideas to add to a portfolio
def findstocks(args:StrategyConfig, verbose:bool=False, sell:bool=False, skip_tickers:list=None):
    """
    Find the stocks based on the strategy configuration provided.

    Args:
        args (StrategyConfig): An object that contains the strategy configuration.
        verbose (bool, optional): Whether to print verbose output. Default is False.
        sell (bool, optional): Whether to find stocks to sell. Default is False.
        skip_tickers (list, optional): A list of stock tickers to skip. Default is None.
    Returns:
        None

    Prints:
        If verbose is True, it prints the answer, stock list, and prediction report.

    Raises:
        None
    """    
    stklist, answer = stock_research(sell, skip_tickers)
    portfolio_path = os.path.join(ROOT_D,'tmp.csv')
    with open(portfolio_path, 'w') as f:
        f.write('Description,Ticker,Quantity,Price,Cost,Unit Cost,Value,Gain,Gain %\n')
        for stk in stklist:
            price = get_price(stk)
            f.write(f'{stk},{stk},1,{price},{price},{price},0,0,0\n')

    if verbose:
        print(answer)
        print(stklist)
        if sell:
            print('SELL PREDICTION REPORT'+20*'=')
        else: 
            print('BUY PREDICTION REPORT'+20*'=')
    # run backtest to evaluate the portfolio-----------------------
    backtester = Backtester(portfolio_path, args=args, zero=True)    
    future_days = [30, 120, 360]
    pred_result = backtester.rank_predict(args.present_date, future_days)
    if sell:
        gen_report(backtester, pred_result, 'findsell', report=answer)
    else:
        gen_report(backtester, pred_result, 'find', report=answer)
    
    if args.run_trade and sell==False:
        action = backtester.strategy.generate_allocation()
        alpaca_execute(action)

    return backtester
   
def stock_research(sell=False, skip_tickers:list=None):
    """
    Perform stock research and provide a list of the best 10 growth stocks in the information technology sector.
    Args:
        sell (bool): If True, search for stocks to sell. Default is False.
        skip_tickers (list): List of stock tickers to skip. Default is None.
    Returns:
        tuple: A tuple containing the list of tickers for the recommended stocks and the research answer.

    Raises:
        Any exceptions that may occur during the execution.
    """    
    cur_date = datetime.now().strftime(DATE_FORMAT)
    # get news to search for stocks
    if sell:
        sysPrompt = 'You are a financial advisor. Given today is {}, ' \
            'respond with a list of the worst 10 stocks in the {} sector.' \
            'List {} stocks that is expected to decline their stock price in a year in the future.'
        if skip_tickers is not None:
            skip_ = ', '.join(skip_tickers)
            sysPrompt += f' In your analysis, avoid the following stocks: {skip_}'
    else:
        sysPrompt = 'You are a financial advisor. Given today is {}, ' \
            'respond with a list of the best 10 growth stocks in the {} sector.' \
            'List {} stocks that is expected to grow their stock price in a year in the future.'
        
    query = sysPrompt.format(cur_date, SECTOR, MARKET_CAP)
    answer = asyncio.run(run_gptresearch(query))
    p = get_tickers(answer)
    return p, answer


def get_args(raw_args=None):
    """
    Parse and return the command-line arguments.

    Args:
        raw_args (list): A list of strings representing the command-line arguments. Default is None.

    Returns:
        argparse.Namespace: An object containing the parsed arguments.

    Description:
        This function uses the ArgumentParser class to define and parse command-line arguments. The function accepts the following arguments:
        --strategy (-s): The strategy to use for prediction. Default is 'GPTStrategy'.
        --inference_days (-i): Number of days for inference. Default is 20.
        --future_days (-f): Number of days in the future to predict. Default is 10.
        --newsapi (-n): The source of news articles. Options are 'finhub', 'newsapi', or 'google'. Default is 'finhub'.

    Example:
        To parse command-line arguments:
        ```
        args = get_args()
        print(args.strategy)
        print(args.inference_days)
        print(args.future_days)
        print(args.newsapi)
        ```
    """    
    parser = ArgumentParser(description="parameters")
    parser.add_argument('--strategy','-s', type=str, default='GPTStrategy', help=f'Modes: {__MODEL_REGISTRY__.keys()}')
    parser.add_argument('--inference_days','-i', type=int, default=20, help='number of days for inference')
    parser.add_argument('--future_days','-f', type=int, default=10, help='number of days in the future to predict')
    parser.add_argument('--newsapi','-n', type=str, default="finhub", help='newsapi: finhub, newsapi, google')
    args = parser.parse_args(raw_args)
    return args


if __name__ == "__main__":
    args = get_args()
    print(args)
    assert args.strategy != 'ManualStrategy', f"Invalid strategy: {args.strategy}"
    config = StrategyConfig(
        inference_days=args.inference_days,
        future_days=args.future_days,
        strategy=args.strategy,
        newsapi=args.newsapi
    )
    config.present_date = date.today()
    config.past_date = find_business_day(config.present_date, -config.inference_days)
    findstocks(config)
    

    
    
    

    


# integration with alpaca API

from alpaca.trading.requests import GetAssetsRequest
from alpaca.trading.enums import AssetClass
from alpaca.trading.requests import MarketOrderRequest
from alpaca.trading.enums import OrderSide, TimeInForce
from alpaca.trading.requests import LimitOrderRequest
from alpaca.trading.client import TradingClient
import pandas as pd
from atradebot.params import ALPACA_API, ALPACA_SECRET, PAPERTRADE
from atradebot.utils.portfolio import Portfolio, dict2portfolio
from atradebot.utils.utils import find_business_day
from datetime import date
from atradebot.strategies.strategy import StrategyConfig

ALPACA_CLIENT = TradingClient(ALPACA_API,ALPACA_SECRET,paper=PAPERTRADE)

def _buy(stock, allocation):
    market_order_data = MarketOrderRequest(
                        symbol=stock,
                        side=OrderSide.BUY,
                        qty=allocation,
                        time_in_force=TimeInForce.DAY,
                        )
    market_order = ALPACA_CLIENT.submit_order(order_data=market_order_data)
    return market_order

def _sell(stock, allocation):
    market_order_data = MarketOrderRequest(
                        symbol=stock,
                        side=OrderSide.SELL,
                        qty=allocation,
                        time_in_force=TimeInForce.DAY,
                        )
    market_order = ALPACA_CLIENT.submit_order(order_data=market_order_data)
    return market_order

def alpaca_execute(allocation):
    """
    Execute trades based on the given allocation using Alpaca API.
    Args:
        allocation (dict): A dictionary containing stock tickers as keys and their respective allocations as values.
    Returns:
        None    
    """
    for stock, allocation in allocation.items():
        if allocation > 0:
            _buy(stock, allocation)
        elif allocation < 0:
            _sell(stock, -allocation)

def alpaca_get_portfolio():
    """
    Get the current portfolio from Alpaca.

    Returns:
        Portfolio: An object representing the portfolio containing stock tickers and their respective allocations.
    """
    positions = ALPACA_CLIENT.get_all_positions()
    dict_port = {}
    for position in positions:
        if position.qty != 0:
            dict_port[position.symbol] = (float(position.qty), float(position.cost_basis)) # tuple (quantity, cost basis)

    portfolio_path = dict2portfolio()
    present_date = date.today()
    past_date = find_business_day(present_date, -30)
    args = StrategyConfig(
        past_date=past_date,
        present_date=present_date,
    )
    portfolio = Portfolio(portfolio_path, args)
    return portfolio


# Description: GPT-based strategies

import torch
import re
from openai import OpenAI
from atradebot.params import OPENAI_API, DATE_FORMAT, ALLTICKS
from datetime import datetime
from atradebot.data.yf_data import get_yf_data_some, get_yf_data_all
from atradebot.data.llm_prompt import LLM_INSTRUCTION, construct_prompt
from atradebot.utils.utils import find_business_day, extract_last_number
from atradebot.strategies import register_strategy
from atradebot.strategies.strategy import Strategy, StrategyConfig

OPENAI_CLIENT = OpenAI(api_key=OPENAI_API)

@register_strategy("GPTStrategy")
class GPTStrategy(Strategy): 
    """
    A class representing a GPTStrategy for generating stock allocation.

    Attributes:
        inference_days (int): The number of days used for prediction inference.
        future_days (int): The number of days into the future to predict.
        newsapi (NewsAPI): The API used for fetching news data.
        client (OpenAI): The OpenAI client for accessing the GPT model.
    """    
    def __init__(self, args:StrategyConfig):
        """
        Initialize the Strategy object with the specified configuration.

        Args:
            args (StrategyConfig): An object containing the configuration settings.

        Returns:
            None

        Attributes:
            inference_days (int): Number of days to use for inference.
            future_days (int): Number of days to predict into the future.
            newsapi (str): API key for accessing news data.
            client (OpenAI): Client object for accessing OpenAI API.
        """        
        self.inference_days = args.inference_days
        self.future_days = args.future_days
        self.newsapi = args.newsapi
        self.client = OPENAI_CLIENT
        # NOTE: the model may already know the current stock price, so cant backtest it
    
    def generate_allocation(self, present_date:datetime.date, portfolio):
        """
        Generate allocation for the given portfolio based on future price predictions.

        Args:
            present_date (datetime.date): The date for which allocation is to be generated.
            portfolio: An object representing the portfolio containing tickers.

        Returns:
            dict: A dictionary with stock tickers as keys and allocation quantity as values.
        """        
        future_price, allocation = {}, {}
        for stock in portfolio.tickers:
            _, pred_pct, _ = self.predict(present_date, self.future_days, stock)
            future_price[stock] = pred_pct
        # buy top5 and sell bottom5
        top5 = sorted(future_price.items(), key=lambda x: x[1], reverse=True)[:5]
        bot5 = sorted(future_price.items(), key=lambda x: x[1], reverse=False)[:5]
        buy_qnt = 5; sell_qnt = 5
        for stock, pred in top5:
            if pred > 0:
                allocation[stock] = buy_qnt
                buy_qnt -= 1
        for stock, pred in bot5:
            if pred < 0:
                allocation[stock] = -sell_qnt
                sell_qnt -= 1
        return allocation
    
    def predict(self, present_date:datetime.date, future_days:int, stock:str): 
        """
        Predict the future stock price based on past data and GPT-4o model.

        Args:
            present_date (datetime.date): The current date for which the prediction is being made.
            future_days (int): Number of days in the future to predict the stock price.
            stock (str): The stock symbol for which the prediction is being made.

        Returns:
            tuple: A tuple containing the log dictionary, predicted percentage change, and predicted future stock price.

        Raises:
            None
        """        
        past_date = find_business_day(present_date, -self.inference_days)
        stock_data = get_yf_data_some(stock, past_date, present_date)
        prompt, news = construct_prompt(stock, present_date, past_date, futuredays=future_days, newsapi=self.newsapi)

        completion = self.client.chat.completions.create(
                model="gpt-4o",
                messages=[{"role": "system", "content": LLM_INSTRUCTION},
                    {"role": "user", "content": prompt}])
        message = completion.choices[0].message.content
        answer = message.split('%')[0]
        pred = extract_last_number(answer)
        torch.cuda.empty_cache()
        cur_price = stock_data['Close'].iloc[-1, -1]
        if pred is None:
            print("Failed to get prediction from the model.")
            return None, 0, cur_price
        
        price_ = cur_price*(pred/100) + cur_price
        log = {"prompt": prompt, "answer": answer, "news": news.to_dict()}
        return log, pred, price_  
    
if __name__ == "__main__":
    args = StrategyConfig()
    f = GPTStrategy(args)
    present_date = datetime.strptime("2024-12-10", DATE_FORMAT)
    a, pred, price = f.predict(present_date, 10, "AAPL")
    print(pred, price)


# technical analysis functions
import pandas as pd
import numpy as np
import os
import matplotlib.pyplot as plt
from scipy import stats
from argparse import ArgumentParser
from datetime import datetime, date
import random
from atradebot.params import DATE_FORMAT, ROOT_D
from atradebot.utils.utils import find_business_day, get_datetime
from atradebot.data.yf_data import get_yf_data_some, get_price


class Portfolio():
    """
    A class representing a portfolio.

    Attributes:
        portfolio_path (str): The path to the portfolio csv file.
        past_date (datetime.date): The start date for price data collection (past).
        present_date (datetime.date): The end date for price data collection.
        outdir (str): The output directory path.
    """    
    def __init__(self, 
                 portfolio_path:str, 
                 args,
                 outdir='./') -> None:
        
        '''
        Args:
            portfolio_path (str): path to the portfolio csv file
            past_date (datetime.date): start of date for price data collect (past)
            present_date (datetime.date): end of date for price data collect
        '''
        past_date = get_datetime(args.past_date)
        present_date = get_datetime(args.present_date)

        # load portfolio
        self.outdir = outdir
        self.infile = portfolio_path
        # pd.DataFrame columns: ['Description', 'Ticker', 'Quantity', 'Price', 'Cost', 'Unit Cost', 'Value', 'Gain', 'Gain %']
        
        stocks = pd.read_csv(self.infile)
        stocks_lst = stocks[stocks['Quantity'].notnull()]
        stocks_lst = stocks_lst.dropna()
        stocks_lst['Ticker'] = stocks_lst['Ticker'].str.strip()
        self.tickers = stocks_lst['Ticker'].tolist()
        self.portfolio = stocks_lst # current portfolio
        cols2float = ['Price', 'Cost', 'Unit Cost', 'Value', 'Gain', 'Gain %']
        self.portfolio[cols2float] = self.portfolio[cols2float].astype(float)

        # Get price DATA
        self.today = find_business_day(present_date)
        self.pastday = find_business_day(past_date)
        self.evalday = self.today # update_portfolio will change the portfolio evaluation date
        self.get_data(self.pastday, self.today) 
        # sets: self.today, self.pastday, self.price_data, self.benchmark_price

        self.cash_available = args.cash # cash available for trading

    def __str__(self):
        """
        Return a string representation of the object.

        This method calculates and prints the portfolio beta, alpha and value.
        It also returns a string representation of the portfolio.

        Returns:
            str: A string representation of the portfolio.
        """        
        portfolio_value = self.get_portfolio_value()
        beta, alpha = self.calc_beta()
        print("The portfolio beta is ", round(beta, 4)," alpha is ", round(alpha,5))
        print(f"The portfolio value is {portfolio_value}")
        return self.portfolio.to_string()
    
    def get_cost(self):
        """
        Get portfolio cost.

        Returns:
            int: Total portfolio cost.
        """        
        return self.portfolio['Cost'].sum()
    
    def get_portfolio_value(self):
        """
        Get portfolio value.

        Returns:
            int: Total portfolio value.
        """        
        return self.portfolio['Value'].sum()

    def get_data(self, pastday:datetime.date, presentday:datetime.date):
        """
        Get price data for a list of tickers within a specified date range.

        Args:
            pastday (datetime.date): The start date for the data retrieval.
            presentday (datetime.date): The end date for the data retrieval.

        Returns:
            None

        Side Effects:
            - Updates the 'price_data' attribute of the object with fetched data for the tickers.
            - Drops any rows with missing values from the 'price_data'.
            - Updates the 'benchmark_price' attribute with data fetched for the S&P 500 index (ticker 'SPY').
            - Calls the 'update_portfolio' method to update the portfolio based on the retrieved data.
        """        
        self.price_data = get_yf_data_some(self.tickers, past_date=pastday, present_date=presentday) 
        self.price_data = self.price_data.dropna()
        self.benchmark_price = get_yf_data_some('SPY', past_date=pastday, present_date=presentday)
        self.update_portfolio()

    def zero_allocation(self):
        """
        Zero out the allocation values in the portfolio.

        This function sets all allocation values in the portfolio to zero.

        Args:
            self: The object instance representing the portfolio.

        Returns:
            None
        """        
        self.portfolio['Quantity'] = 0
        self.portfolio['Cost'] = 0
        self.portfolio['Unit Cost'] = 0
        self.portfolio['Value'] = 0
        self.portfolio['Gain'] = 0
        self.portfolio['Gain %'] = 0

# Technical Analysis --------------------------------------------
    def calc_beta(self):
        """
        Calculate beta and alpha of a portfolio against the S&P 500.

        Returns:
            tuple: beta, alpha
        """    
        cost = self.portfolio['Cost'].tolist()
        weights = cost/np.sum(cost)

        price_data_adj = self.price_data['Close']
        ret_data = price_data_adj.pct_change()[1:]
        port_ret = (ret_data * weights).sum(axis = 1)
        
        benchmark_ret = self.benchmark_price["Close"].pct_change()[1:]
        b_v = np.squeeze(benchmark_ret)
        (beta, alpha) = stats.linregress(b_v, port_ret.values)[0:2]

        return beta, alpha
    
# Flags ---------------------------------------------------------
    def stop_loss(self, loss=0.4):
        # check if the stock goes down loss_percent from the cost basis 
        """
        Calculate the stop loss levels for the portfolio.

        Args:
            loss (float): The percentage loss at which the stop loss should trigger. Default is 0.4.

        Returns:
            dict: A dictionary containing information on the assets in the portfolio that have reached the stop loss level.

        Raises:
            None
        """        
        stop_loss = self.portfolio['Unit Cost'] * (1 - loss)
        flag = self.portfolio['Price'] < stop_loss
        return self.portfolio[flag]
    
    def stop_down(self, days=5):
        # check if the stock is going down for days consecutively 
        """
        Calculate the percentage change in exponential moving average (EMA) with respect to the maximum EMA value over a specified number of days.

        Args:
            self: The object instance.
            days (int): Number of days to consider for the EMA calculation. Default is 5.

        Returns:
            pandas.Series: A pandas Series containing the percentage change in EMA values sorted in ascending order.
        """        
        data_ema = self.price_data['Close'].ewm(span=days, adjust=False).mean() # EMA
        data_ema_max = data_ema.max()
        # get ema percentage change for the last days
        ema_pct_change = (data_ema.iloc[-1]-data_ema_max)/data_ema.iloc[-1]
        ema_pct_change.sort_values(ascending=True, inplace=True)
        return ema_pct_change

# Portfolio management -----------------------------------------
    def buy_stock(self, stock:str, quantity:int):
        # add stock to the portfolio
        """
        Buy a stock and update the portfolio with the transaction details.

        Args:
            stock (str): The stock symbol to buy.
            quantity (int): The quantity of the stock to buy.

        Returns:
            None

        Raises:
            None
        """        
        day = self.evalday.strftime(DATE_FORMAT)
        if stock in self.tickers: # add to existing stock
            idx = self.portfolio[self.portfolio['Ticker'] == stock].index
            price = self.portfolio.loc[idx, 'Price']
            self.portfolio.loc[idx, 'Quantity'] += quantity
            self.portfolio.loc[idx, 'Cost'] += price * quantity
            self.portfolio.loc[idx, 'Value'] += price*quantity
            self.portfolio.loc[idx, 'Gain'] = self.portfolio['Value'][idx] - self.portfolio['Cost'][idx]
            self.portfolio.loc[idx, 'Gain %'] = self.portfolio['Gain'][idx]/self.portfolio['Cost'][idx]
            self.portfolio.loc[idx, 'Unit Cost'] = self.portfolio['Cost'][idx]/self.portfolio['Quantity'][idx]
        else: # new stock append to portfolio
            new_prices = get_yf_data_some(stock, past_date=self.pastday, present_date=self.today)
            price = new_prices['Close'].loc[day]
            self.price_data =  pd.concat([self.price_data, new_prices],axis=1)
            self.tickers.append(stock)
            cost = price * quantity
            new_stock = pd.DataFrame({'Description': stock, 'Ticker': stock, 'Quantity': quantity, 
                                    'Price': price, 'Cost': cost, 'Unit Cost': cost/quantity, 
                                    'Value': price*quantity, 'Gain': price*quantity - cost, 'Gain %': (price*quantity - cost)/cost})
            self.portfolio = pd.concat([self.portfolio, new_stock])
    
    def sell_stock(self, stock:str, quantity:int):
        # sell stock from the portfolio
        """
        Sell a specific stock from the portfolio.

        Args:
            stock (str): The ticker symbol of the stock to be sold.
            quantity (int): The quantity of the stock to be sold.

        Returns:
            float: The total revenue generated from selling the stock.

        Raises:
            None
        """        
        idx = self.portfolio[self.portfolio['Ticker'] == stock].index
        if idx.empty:
            # print(f"Stock {stock} not found in portfolio")
            return 0
        revenue = quantity * self.portfolio.loc[idx, 'Price']
        if quantity >= self.portfolio['Quantity'][idx[-1]]: # sell all
            self.portfolio = self.portfolio.drop(idx)
            self.tickers.remove(stock)
            self.price_data.drop(stock, axis=1, inplace=True)
        else:
            self.portfolio.loc[idx,'Quantity'] -= quantity
            self.portfolio.loc[idx,'Cost'] -= self.portfolio['Unit Cost'][idx]*quantity
            self.portfolio.loc[idx,'Value'] -= self.portfolio['Price'][idx]*quantity
            self.portfolio.loc[idx,'Gain'] = self.portfolio['Value'][idx] - self.portfolio['Cost'][idx]
            self.portfolio.loc[idx,'Gain %'] = self.portfolio['Gain'][idx]/self.portfolio['Cost'][idx]
        return revenue
    
    def update_portfolio(self, day:datetime.date=None):
        # update portfolio
        """
        Update the portfolio with the latest price data.

        Args:
            day (datetime.date, optional): The date of evaluation. If None, the latest available date will be used.

        Returns:
            None

        Updates:
            - 'Price', 'Value', 'Gain', and 'Gain %' columns in the portfolio DataFrame.

        Raises:
            KeyError: If the close price data for the specified day is not available.
        """        
        if day:
            self.evalday = day
            day = day.strftime(DATE_FORMAT)
            d = self.price_data['Close'].loc[day].to_dict()
        else:
            d = self.price_data['Close'].iloc[-1].to_dict()
        self.portfolio['Price'] = self.portfolio["Ticker"].map(d).fillna(self.portfolio["Price"])
        self.portfolio['Value'] = self.portfolio['Price'] * self.portfolio['Quantity']
        self.portfolio['Gain'] = self.portfolio['Price'] * self.portfolio['Quantity'] - self.portfolio['Cost']
        self.portfolio['Gain %'] = (self.portfolio['Price'] * self.portfolio['Quantity'] - self.portfolio['Cost']) / self.portfolio['Cost']

    def save_back(self):
        """
        Save the portfolio data to a CSV file.

        Creates the output directory if it does not exist and saves the portfolio data to a CSV file.

        Raises:
            OSError: If the output directory cannot be created or the CSV file cannot be saved.
        """        
        if not os.path.exists(self.outdir):
            os.mkdir(self.outdir)
        outfile = os.path.join(self.output_dir,'portfolio', datetime.now().strftime("%b%d_%H-%M-%S"), '.csv')
        self.portfolio.to_csv(outfile, index=False)

    def plot_portfolio(self):
        # plot portfolio
        """
        Plot the portfolio values for each ticker.

        This method plots the prices for each ticker in the portfolio in a bar chart.

        Args:
            self: The Portfolio object instance.

        Returns:
            None

        Note:
            This method requires the 'Ticker' and 'Price' columns in the portfolio DataFrame.

        Example:
            portfolio.plot_portfolio()
        """        
        self.portfolio.plot(x='Ticker', y='Price', kind='bar')
        plt.show()

    def execute(self, action:dict):
        # execute
        act = False
        for stock, qnt in action.items():
            # buy
            if qnt > 0:
                self.buy_stock(stock, qnt)
                act = True
            # sell
            elif qnt < 0:
                self.sell_stock(stock, qnt)
                act = True
        return act


def dict2portfolio(portfolio:dict):
    """
    Convert a dictionary to a portfolio and save it to a CSV file.

    Args:
        portfolio (dict): A dictionary containing stock tickers and their (respective quantities, cost basis).
        outfile (str): The output file path for the CSV file.

    Returns:
        str: The file path of the created portfolio.

    Raises:
        None.
    """    
    portfolio_path = f'{ROOT_D}/sd/tmp_.csv'
    with open(portfolio_path, 'w') as f:
        f.write('Description,Ticker,Quantity,Price,Cost,Unit Cost,Value,Gain,Gain %\n')
        for stk in portfolio.keys():
            price = get_price(stk)
            qnt = portfolio[stk][0]
            cost = portfolio[stk][1]
            f.write(f'{stk},{stk},{qnt},{qnt*price},{cost},{cost/qnt},{qnt*price},{qnt*price-cost},{(qnt*price-cost)/cost}\n')
    return portfolio_path


def test_portfolio(stklist:list, rnd=False):
    """
    Create a test portfolio based on a list of stocks.

    Args:
        stklist (list): A list of stock tickers.
        rnd (bool): If True, generate random quantities for the stocks. Default is False.

    Returns:
        str: The file path of the created portfolio.

    Raises:
        None.
    """    
    portfolio_path = f'{ROOT_D}/sd/tmp.csv'
    with open(portfolio_path, 'w') as f:
        f.write('Description,Ticker,Quantity,Price,Cost,Unit Cost,Value,Gain,Gain %\n')
        for stk in stklist:
            price = get_price(stk)
            if rnd:
                qnt = random.randint(1, 5)
            else:
                qnt = 0
            f.write(f'{stk},{stk},{qnt},{qnt*price},{qnt*price},{price},0,0,0\n')
    return portfolio_path

if __name__ == "__main__":
    portfolio_path = f'{ROOT_D}/sd/test_portfolio.csv'
    portfolio = Portfolio(portfolio_path, past_date='2023-04-05', present_date='2023-08-13')
    beta, alpha = portfolio.calc_beta()
    print("The portfolio beta is", round(beta, 4))
    print("The portfolio alpha is", round(alpha,5))
    bdays = 3
    sloss = 0.4
    print(f"The stocks going down past {bdays} days: ", portfolio.stop_down(bdays))
    print(f"The stocks went down {sloss} from cost basis are: ", portfolio.stop_loss(loss=sloss))
    
    portfolio.update_portfolio(day=get_datetime('2023-08-07'))
    print(portfolio)
    portfolio.buy_stock('AAPL', 10)
    print(portfolio)
    portfolio.sell_stock('AAPL', 5)
    print(portfolio)
    


    


# create dataset for training a financial model on Yahoo Finance data API, Google News
# collect news data for stocks during valley and peak days for a stock

# prompt creation: news + stock_data + date -> prompt 

import os
from argparse import ArgumentParser
import datasets
from datasets import Dataset
from atradebot.data.yf_data import get_yf_data_some, get_yf_detail
from atradebot.utils.utils import find_peaks_valleys, filter_points
from atradebot.params import DATE_FORMAT
from atradebot.utils.utils import find_business_day
from datetime import datetime
import pandas as pd
from atradebot.data.get_news import get_news


LLM_INSTRUCTION = """You are a seasoned stock market analyst. \
    You are given past news and basic financials from companies. \
    Your task is to provide prediction for the companies' stock price movement and forecast possible future news. 
    Your answer format should be as follows:\n\n {'Price Change Future Prediction': price percent_change (%), 'Future news': text}\n"""
    
def format_chat_template(tokenizer, row): 
    """
    Format a chat template using a given row of data.

    Args:
        tokenizer: An object that handles text tokenization.
        row (dict): A dictionary containing 'instruction' and 'response' keys.

    Returns:
        dict: The updated row with 'text' key containing the formatted chat template.
    """    
    row_json = [{"role": "system", "content": LLM_INSTRUCTION },
                {"role": "user", "content": row["instruction"]},]
    if 'response' in row:    
        row_json += [{"role": "assistant", "content": row["response"]}]
    
    row["text"] = tokenizer.apply_chat_template(row_json, tokenize=False)
    return row

def construct_prompt(stock:str, 
                     present_date:datetime.date, 
                     past_date:datetime.date, 
                     futuredays:int=5, 
                     newsapi:str='dataset',
                     num_results = 5):
    """
    Construct a prompt for a stock based on historical news data.

    Args:
        stock (str): The stock symbol for which the prompt is being constructed.
        present_date (datetime.date): The present date.
        past_date (datetime.date): The past date to retrieve news from.
        futuredays (int, optional): Number of future days to consider for the prompt. Default is 5.
        newsapi (str, optional): Source of news data. Default is 'dataset'.

    Returns:
        tuple: A tuple containing the constructed prompt and a DataFrame of news data.

    Raises:
        None
    """    
    timeperiod = [past_date.strftime(DATE_FORMAT), present_date.strftime(DATE_FORMAT)]
    news = get_news(stock, timeperiod, num_results=num_results, news_source=newsapi)
    if len(news) == 0: # if cant find news, use random past dates from dataset news
        news = get_news(stock, [], num_results=num_results, news_source='dataset')

    news = pd.DataFrame(news)

    # prune to keep prompt short (avoid OOM)
    # if len(news) > num_results: # truncate to 5 news
    #     news = news.sample(num_results)
    # news = news.drop_duplicates(subset=['Date'])# remove repeated date news

    now = True if present_date == datetime.today().date() else False
    prompt = build_context(news, stock, futuredays, now=now)
    # print('Prompt context len: ', len(prompt), len(news))
    return prompt, news

def collect_events(stock:str, past_date:str, present_date:str, ret=False):
    """
    Collect events to gather news.

    Args:
        stock (str): Stock id.
        past_date (str): Date start to collect data in format yyyy-mm-dd (past).
        present_date (str): Date end to collect data in format yyyy-mm-dd (today).
        ret (bool, optional): If True, return random events, defaults to False.

    Returns:
        list[pandas.Timestamp]: List of dates to collect news.
    """    
    
    data = get_yf_data_some(stock, past_date=past_date, present_date=present_date)
    if data.empty or ret: # generate random events if no data
        l = data.index.to_list()
        events = [ll for i, ll in enumerate(l) if i%2]
        return events

    data_mean = []
    data_stock = data['Close'].values.squeeze()
    for i in range(len(data_stock)-3):
        data_mean.append(data_stock[i:i+3].mean())

    # collect when to gather news
    p, v = find_peaks_valleys(data_mean)
    # index of data_mean that are peaks and valleys
    peak_idx, valley_idx = filter_points(data_mean, p, v) 
    events = peak_idx + valley_idx #concat lists
    events += [2] # add second day as first event to collect news
    events.sort()
    events_dates = [data.index[event] for event in events]
    return events_dates, data


def build_context(news, stock, futuredays, now=False):
    """
    Build a prompt with news about a specific stock and request for the price percentage change after a certain number of days.

    Args:
        news (DataFrame): A DataFrame containing news data with columns 'Date' and 'News'.
        stock (str): The name of the stock.
        futuredays (int): The number of days in the future to predict the stock price change.
        now (bool, optional): If True, include today's information. Defaults to False.

    Returns:
        str: A prompt with news about the stock and a request for the price percentage change after a certain number of days.
    """   

    # Add stock description and company info
    company, company_text, company_person, company_info = get_yf_detail(stock)
    prompt = f"{stock}: {company_text} \n"
    if now: # can add today's info if not using for backtesting
        prompt += f"{company_person} \n Company Details: {company_info} \n"
    
    # Add News
    if len(news) > 0:
        prompt += f"Here is a list of news about {stock}: \n"
        
        news_grouped = news.groupby('Date')['News'].apply(' '.join).reset_index() # concatenate news for same date
        for _, row in news_grouped.iterrows():
            prompt += f"Date: {row['Date']}, News: {row['News']} \n"

    # Add instruction
    prompt += f"Please, provide the price percentage change (%) for the {stock} stock price after {futuredays} days in the future: \n"
    return prompt


def create_prompt(dataset:Dataset, stock, backdays, futuredays, getevent=False):
    # for llama3
    """
    Create a prompt dataset for a given stock.

    Args:
        dataset (Dataset): The dataset containing stock data.
        stock (str): The stock ticker symbol to create prompt for.
        backdays (int): Number of days to look back.
        futuredays (int): Number of days to look into the future.
        getevent (bool, optional): Flag to determine if events should be collected. Defaults to False.

    Returns:
        Dataset: A dataset containing prompts and responses for the stock.

    Raises:
        None
    """    
    dhf = dataset.to_pandas()
    # filter for a stock
    dhf = dhf[dhf['Tickers'] == stock]

    # get max and min dates in dataset 
    max_date = max(dhf['Date'])
    min_date = min(dhf['Date'])
    # gather interesting dates
    if getevent:
        events, data = collect_events(stock, min_date, max_date)
    else:
        events = None
        data = get_yf_data_some(stock, min_date, max_date)
    data = data.reset_index() # reset index to get date

    if not events:
        # get random dates between min and max date
        events = data['Date'].sample(int(len(data['Date']) * 0.7))

    # collect group of news and financials for that stock up to some dates behind (backdays) the interesting date
    llm_dataset = [] # TODO need streaming 
    for event in events:
        past = find_business_day(event, -backdays)
        if past < min_date:
            continue
        future = find_business_day(event, futuredays)
        if future > max_date:
            continue
        # get news and financials for that stock up to backdays behind the interesting date
        # news = dhf.filter(lambda x: x['Date'] <= event and x['Date'] >= past)
        news = dhf[(dhf['Date'] <= event) & (dhf['Date'] >= past)]
        if len(news) == 0:
            continue
        # TODO: add some financials
        # create prompt
        prompt = build_context(news, stock, futuredays)
        # create response
        # news_future = dhf.filter(lambda x: x['Date'] == future)
        news_future = dhf[dhf['Date'] == future]
        if len(news_future) == 0:
            continue
        resp = ""
        news_grouped = news_future.groupby('Date')['News'].apply(' '.join).reset_index()
        for _, row in news_grouped.iterrows():
            resp += f"Date: {row['Date']}, News: {row['News']} \n"
  
        fval = data.loc[data['Date'] == future, 'Close']
        if len(fval) == 0:
            continue
        
        pred = fval.values[0] - data.loc[data['Date'] == event, 'Close'].values[0]
        pred = pred / data.loc[data['Date'] == event, 'Close'].values[0] * 100
        response = {"Prediction": pred[0], "Future news": resp}
        llm_dataset.append({'instruction': prompt, 'response': response})

    datas = Dataset.from_list(llm_dataset)
    return datas

if __name__ == "__main__":
    pass



# collect news via api every day. Use with cron job. 
# crontab -e
# 30 2 * * * collect.py
from atradebot.data.get_news import get_news
import random
from datetime import datetime, timedelta, date
from atradebot.params import DATE_FORMAT, HF_LOCALDATA
import datasets
import os
from atradebot.utils.search import stock_research
import json


if __name__ == "__main__": 
    # select some stocks to gather news
    stocks, answ = stock_research()
    today = date.today()
    past = today - timedelta(days=2)

    month = today.strftime('%m')
    year = today.strftime('%Y')
    HF_LOCALDATA += f'{month}_{year}.jsonl'
    print(f'{today} Collecting news for: {stocks}')
    news_text = [{"link": '',
                        "title": '',
                        "snippet": '',
                        "date": today.strftime(DATE_FORMAT),
                        "source": 'gptresearch',
                        "text": answ,
                        "stock": ' '.join(stocks)}]
    for stock in stocks:
        news = get_news(stock, 
                time_period=[past.strftime(DATE_FORMAT), today.strftime(DATE_FORMAT)], 
                num_results=5, 
                news_source='finhub')
        news_text.extend(news)

    print('Collected news:', len(news_text))
    with open(HF_LOCALDATA, "a") as file:  # Open in append mode
        for entry in news_text:
            file.write(json.dumps(entry) + "\n")

    print('All done!')
